{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name colors}
{@description Color processing}

{$I directives.pas}
uses simage,getparam;
var par              : tparams;
    cc               : CSImg_Config;
    img              : TSImg_FPImage;
    fi,fo            : string;
    ca,cad,fd        : boolean;
    s1,s2,ms,cas     : double;
    f                : array[1..9] of double;
    i,c              : longint;
    d                : ^simgfp;

begin
    with par do
    begin
        init;
        paramusage:='[<infile> [<outfile>]]';
        fi:='';      defparam('<infile>','Source image (default: stdin)',fi);
        fo:='';      defparam('<outfile>','Destination image (default: stdout)',fo);
                     defparam('-h','Help',helpf);
        ca:=false;   defparam('-ca','Adjust colors',ca);
        cad:=false;  defparam('-cad','Use differentiation mode for color adjustment',cad);
        cas:=1.5;    defparam('-cas','Smoothing radius for color adjustment algorithm',cas);
        s1:=1;       defparam('-s1','Saturation method 1 (default: 1)',s1);
        s2:=1;       defparam('-s2','Saturation method 2 (default: 1)',s2);
        ms:=2;       defparam('-ms','Maximum saturation for first algorithm (default: 2)',ms);
        for i:=1 to 9 do
        begin
            f[i]:=1.0;
            defparam('-f'+int2str(i),'Manual color correction factor for color '+int2str(i),f[i]);
        end;
                     defparam('-o','Destination image',fo); deffmt('-o','<outfile>');
        run;
        done;
    end;
    fd:=false;
    for i:=1 to 9 do
        fd:=fd or ((f[i]-1)>0.001);

    cc:=CSImg_Config.create(SImg_Config_Default);
    cc.debug(cc.name+': source='+fi+' dest='+fo+' ca='+boolean2str(ca)+' cad='+boolean2str(cad)+' cas='+real2str(cas)
      +' s1='+real2str(s1)+' s2='+real2str(s2)+' maxsat='+real2str(ms)+' fd='+boolean2str(fd));

    { read input }
    cc.Info(cc.name+': Reading input image '+fi);
    SImg_ReadImageFile(cc,fi,img);
    if ca then cc.Info(cc.name+': Color adjustment factors: '+SImg_ColorAdjust(cc,img,cad,cas));

    { manual color factor adjustment }
    if fd and (img.nc>9) then cc.error(cc.name+': Color factor correction only supported for up to 9 colors');
    if fd then
    begin
        cc.Info(cc.name+': Manual color adjustment');
        d:=img.pt-1;
        for i:=0 to img.nx*img.ny-1 do
        begin
            for c:=1 to img.nc do
                d[c]*=f[c];
            d+=img.nc;
        end;
    end;

    { saturation }
    if abs(s1-1)>1e-3 then
    begin
        cc.Info(cc.name+': Changing saturation (method 1)');
        SImg_Saturation01(cc,img,img,s1,ms);
    end;

    { output }
    if abs(s2-1)>1e-3 then
    begin
        cc.Info(cc.name+': Changing saturation (method 2)');
        SImg_Saturation02(cc,img,img,s2);
    end;

    cc.Info(cc.name+': Writing output file '+fo);
    SImg_WriteImageFile(cc,fo,img);

    SImg_Free(cc,img);
    cc.destroy;
end.
