{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name greyscale}
{@description Draws a gray scale for monitor calibration}

{$I directives.pas}
uses simage,getparam;
var par          : tparams;
    cc           : CSImg_Config;
    img          : TSImg_FPImage;
    fo           : ansistring;
    sx,sy,n,x,y  : longint;
    a,r,g        : double;


function getVal(i:longint):double;
var r,s : double;
begin
    r:=2*i/(n-1);
    if r>1 then s:=2-r
    else s:=r;
    if s>1e-10 then s:=0.5*exp(ln(s)*a);
    if r>1 then s:=1-s;
    if s>1e-10 then s:=exp(ln(s)*g);
    result:=s;
end;


begin
    cc:=CSImg_Config.create(SImg_Config_Default);
    img.pt:=nil;
    with par do
    begin
        init;
        paramusage:='[<outfile>]';
        fo:='';       defparam('<outfile>','Destination image (default: stdout)',fo);
                      defparam('-h','help',helpf);
        sx:=1200;     defparam('-sx','X size (default: 800)',sx);
        sy:=60;       defparam('-sy','Y size (default: 100)',sy);
        n:=20;        defparam('-n','Number of steps (default: 25)',n);
        a:=0.01;      defparam('-a','First step (default: 1%)',a);
        g:=1.14;      defparam('-g','Gamma , default 1.14)',g);
        run;
        done;
    end;

    cc.debug('greyscale: sx='+int2str(sx)+' sy='+int2str(sy)+' n='+int2str(n)+' a='+real2str(a));

    a:=-(ln(2)+ln(a)/g)/ln(0.5*(n-1));
    writeln(a:0:3);
    SImg_New(cc,img,sx,sy,1);
    for x:=0 to sx-1 do
    begin
        r:=getval(x*n div sx);
        for y:=0 to sy-1 do
            img.pt[y*img.sl+x]:=r;
    end;

    cc.textVAlign:=0.5;
    cc.textHAlign:=0.5;
    for x:=0 to n-1 do
    begin
        r:=getval(x);
        if r>0.5 then cc.drawcolor[0]:=r-0.5
        else cc.drawcolor[0]:=r+0.5;
        cc.backcolor[0]:=r;
        SImg_Write8x8(cc, img, round(sx/n*(x+0.5)), (sy+1) div 2, real2str(r*100,3)+'%');
    end;

    SImg_WriteImageFile(cc,fo,img);
    SImg_Free(cc,img);
    cc.destroy;
end.
