{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name interpolate}
{@description Interpolation using a (bit)mask.}

{$I directives.pas}
uses simage,getparam,sysutils;
var par         : TParams;
    cc          : CSImg_Config;
    ri          : TSImg_RDImage;
    fi          : TSImg_FPImage;
    mi          : CSImg_MetaInfo;
    ifn,ofn,mfn : ansistring;
    a,d         : longint;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='[<infile> [<outfile>]]';
        ifn:='';     defparam('<infile>','( bit input image (default: stdin)',ifn);
        ofn:='';     defparam('<outfile>','Output image (default: stdout)',ofn);
        mfn:='';     defparam('-m','Input mask. Default: stdin', mfn);  deffmt('-m','<infile>');
        a:=1;        defparam('-a','Bit mask a that defines where to interpolate. Default 1.', a);
        d:=3;        defparam('-d','Filter radius divider, as larger as less low pass filtering. Disabled if >127. Default: 3,',d);
                     defparam('-o','Destination image',ofn); deffmt('-o','<outfile>');
                     defparam('-h','Help',helpf);
        run;
        done;
    end;
    cc.debug(cc.name+': input=' + ifn + ' output=' + ofn + ' mask=' + mfn + ' a=' + int2str(a)+ ' d=' + int2str(d));
    if d<1 then d:=1;

    a:=a and 255;
    if a=0 then
    begin
        cc.warning(cc.name+': Bitmask is zero. Nothing to do.');
        SImg_Halt;
    end;

    if mfn='' then mfn:='-';
    if ifn='' then ifn:='-';
    if (mfn='-') and (ifn='-') then cc.error(cc.name+': Only one input file can be read from stdin');

    { mask }
    cc.info(cc.name+': Reading mask image '+mfn);
    mi:=CSImg_MetaInfo.create;
    ri.pt:=nil;
    SImg_ReadImageFile(cc,mfn,ri,mi);
    if ri.fmt<>SImg_RawByte then cc.error(cc.name+': 8 bit image expected');

    { input }
    cc.info(cc.name+': Reading input image '+ifn);
    fi.pt:=nil;
    SImg_ReadImageFile(cc,ifn,fi,mi);
    if (fi.nx<>ri.nx) or (fi.ny<>ri.ny) or (fi.nc<>ri.nc) then cc.error(cc.name+': Incompatible image dimensions');

    { interpolation }
    cc.info(cc.name+': Interpolation using ' + int2str(cc.threads) + ' threads');
    SImg_InterpolateMasked(cc, fi, ri, a, d);
    SImg_Free(cc,ri);

    { output }
    cc.info(cc.name+': Writing output image '+ofn);
    SImg_WriteImageFile(cc,ofn,fi,mi);
    SImg_Free(cc,fi);
    mi.destroy;
    cc.destroy;
end.
