{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name rawread}
{@description Read from raw device (e.g. from v4l2 device)}

{$I directives.pas}
uses simage,getparam,compatun;

var par              : tparams;
    cc               : CSImg_Config;
    fi,fo            : shortstring;
    n,fmti,bufs      : longint;
    fmt              : TSImg_RawFormat;
    width,height,col : longint;
    rt               : boolean;
    gr               : CSImg_GetRaw;
    t                : double;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='[<outfile>]';
        fi:='';       defparam('<infile>','Raw source (default: stdout)',fi);
        fo:='';       defparam('<outfile(s)>','Destination image(s) (default: stdout)',fo);
        n:=1;         defparam('-n','Images to save (default: 1)',n);

        width:=768;   defparam('-x','Width (default: 768)',width);
        height:=576;  defparam('-y','Height (default: 576)',height);
        col:=1;       defparam('-c','Number of colors (default: 1)',col);
        fmti:=1;      defparam('-f','Bytes per color, possible values: 1=byte, 2=word, 4=single, 8=double  (default: 1)',fmti);

{        v.adr:=0;     defparam('-adr','AD resolution in ADU/e (default: 0)',v.adr);
        v.sron:=1;    defparam('-sron','Square read out noise (default: 1)',v.sron);
        v.ufl:=10;    defparam('-ufl','Overflow limit (default: 2)',v.ufl);
        v.ofl:=250;   defparam('-ofl','Overflow limit (default: 253)',v.ofl);
        v.bias:=0;    defparam('-bias','Bias (offset which is subtracted from each image)',v.bias);
        v.vbl:=false; defparam('-vbl','Vertical blooming (default: false)',v.vbl);
        v.hbl:=false; defparam('-hbl','Horizontal blooming (default: false)',v.hbl); }

        rt:=false;    defparam('-rt','Use a separate capture thread',rt);
        bufs:=8;      defparam('-bufs','Number of video buffers',bufs);
                      defparam('-o','Destination image(s)',fo); deffmt('-o','<outfile(s)>');

                      defparam('-h','help',helpf);
        run;
        done;
    end;

    cc.debug('rawread: fi='+fi+' fo='+fo+' n='+int2str(n)     +
        ' width='+int2str(width)+' height='+int2str(height)+' col='+int2str(col)+' fmti='+int2str(fmti)+
        ' bufs='+int2str(bufs)+' rt='+boolean2str(rt));


    fmt:=SImg_RawByte; //makes the compiler happy
    case fmti of
        1 : fmt:=SImg_RawByte;
        2 : fmt:=SImg_RawWord;
        4 : fmt:=SImg_RawSingle;
        8 : fmt:=SImg_RawDouble;
        else cc.error('rawread: Invalid format: '+int2str(fmti));
    end;

    gr:=CSImg_FileRead.init(cc, fi, bufs, width,height,col, fmt, rt);

    t:=stimer;
    SImg_RawSave(cc,gr,fo,n);
    writeln(stderr);
    t:=n/(stimer-t);
    if t>1 then writeln(stderr,t:0:2,' fps')
    else writeln(stderr,1/t:0:2,' spf');
    cc.destroy;
end.
