{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name regbounds}
{@description Extrapolate undefined regions at boundaries using other colors.}


{$I directives.pas}
uses simage,getparam,sysutils;


var par         : tparams;
    cc          : CSImg_Config;
    ifn,wfn,ofn : TAAStr;
    b,i         : longint;
    r           : double;
    fi,wi       : array of TSImg_FPImage;
    mi          : array of CSImg_MetaInfo;

begin
{define sdbadd_debug}
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='';
        ifn.size:=0;    defparam('-i','Input image. Can be specified multiple times and must occur triple-wise in combination with -o and -w. ',ifn);           deffmt('-i','<image>');
        wfn.size:=0;    defparam('-w','Weight of input image. Can be specified multiple times and must occur triple-wise in combination with -i and -o ',wfn);  deffmt('-w','<image>');
        ofn.size:=0;    defparam('-o','Output image. Can be specified multiple times and must occur triple-wise in combination with -i and -w.\\'+
                                      'Use "." to disable output for that image',ofn);                                                                          deffmt('-o','<image>');
        b:=100;         defparam('-b','Boundary width. Default: 100',b);
        r:=0.25;        defparam('-r','Maximum regularization factor. Default: 0.25',r);
                        defparam('-h','Help',helpf);
        run;
        done;
    end;
    cc.debug(cc.name+': ifn.size='+int2str(ifn.size)+' wfn.size='+int2str(wfn.size)+' ofn.size='+int2str(ofn.size)+' b='+int2str(b)+' r='+real2str(r));
    if ifn.size<1 then
    begin
        cc.warning(cc.name+': No input files defined. Nothing to do');
        SImg_Halt;
    end;
    if (ifn.size<>wfn.size) or (ifn.size<>ofn.size) then cc.error(cc.name+': Number of input images, number of weight images and number of output images must be equal');

    setlength(fi, ifn.size);
    setlength(wi, ifn.size);
    setlength(mi, ifn.size);
    for i:=0 to ifn.size-1 do
    begin
        fi[i].pt:=nil; wi[i].pt:=nil;
        mi[i]:=CSImg_MetaInfo.create();
        cc.Info(cc.name+': Reading input image ' + int2str(i+1) +' (' + ifn.aa[i] + ')');
        SImg_ReadImageFile(cc,ifn.aa[i],fi[i],mi[i]);
        if (fi[i].nx<>fi[0].nx) or (fi[i].ny<>fi[0].ny) then cc.error(cc.name+': Invalid dimensions of input image ' + int2str(i+1) +' (' + ifn.aa[i] + ')');
        cc.Info(cc.name+': Reading weight image ' + int2str(i+1) + ' (' + wfn.aa[i] + ')');
        SImg_ReadImageFile(cc,wfn.aa[i],wi[i]);
        if (wi[i].nx<>fi[0].nx) or (wi[i].ny<>fi[0].ny) or (wi[i].nc<>fi[i].nc) then cc.error(cc.name+': Invalid dimensions of weight image ' + int2str(i+1) +' (' + wfn.aa[i] + ')');
    end;

    { boundary regularization }
    cc.Info('sdbadd: Boundary regularization');
    SImg_RegBounds(cc, fi,wi, b,r);
    for i:=0 to ifn.size-1 do
        SImg_Free(cc, wi[i]);

    { output }
    for i:=0 to ifn.size-1 do
    begin
        if ofn.aa[i]<>'.' then
        begin
            cc.Info(cc.name+': Writing output image ' + int2str(i+1) +' (' + ofn.aa[i] + ')');
            SImg_WriteImageFile(cc,ofn.aa[i],fi[i],mi[i]);
        end;
        SImg_Free(cc,fi[i]);
        mi[i].destroy;
    end;
end.
