{  simg -- image processing library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{******* SImg_BayerCheckInfo *********************************************}
{*************************************************************************}
{ check Bayer pattern and set number of colors
  return true if Bayer pattern is valid }
function SImg_BayerCheckInfo(var bi:TSImg_BayerInfo):boolean;
var i,j,k,c : longint;
    b       : boolean;
begin
    result:=false;
    c:=0;
    for i:=0 to 1 do
        for j:=0 to 1 do
            if bi.pattern[i,j]>c then c:=bi.pattern[i,j];
    for k:=0 to c do
    begin
        b:=true;
        for i:=0 to 1 do
            for j:=0 to 1 do
                if bi.pattern[i,j]=c then b:=false;
        if b then exit;
     end;
    bi.colors:=c+1;
    result:=true;
end;


{*************************************************************************}
{******* SImg_BayerParsePattern ******************************************}
{*************************************************************************}
// check Bayer info and set number of colors
function SImg_BayerParsePattern(cc:CSImg_Config; s:string7):TSImg_BayerPattern;
var x,y,i,j : longint;
    ch      : char;
label error;
begin
    fillchar(result, sizeof(result), 0);   // makes the compiler happy
    j:=2;
    if (length(s)=5) and (s[3]='/') then j:=3
    else if length(s)<>4 then goto error;

    for y:=0 to 1 do
        for x:=0 to 1 do
        begin
            ch:=s[y*j+x+1];
            case ch of
                'R','r' : i:=0;
                'G','g' : i:=1;
                'B','b' : i:=2;
                'W','w' : i:=3;
                else i:=ord(ch)-49;
            end;
            if (i<0) or (i>3) then goto error;
            result[y,x]:=i;
        end;
    exit;

    error:
    cc.error('SImg_BayerParsePattern: Invalid Bayer pattern string: `'+s+''': expect strings like `1223'', `RGGB'', `12/23'', `RG/GB'' (all equivalent)' );
end;

procedure SImg_BayerParsePattern(cc:CSImg_Config; s:string7; var bi:TSImg_BayerInfo);
var i  : longint;
begin
    for i:=0 to 4 do
        bi.cf[i]:=1.0;
    bi.pattern:=SImg_BayerParsePattern(cc,s);
    if not SImg_BayerCheckInfo(bi) then cc.error('SImg_BayerParsePattern: Invalid Bayer pattern');
end;


{*************************************************************************}
{******* SImg_BayerFlipattern ********************************************}
{*************************************************************************}
function SImg_BayerFlipPattern(cc:CSImg_Config; flipx,flipy:boolean; s:string7):string7;
var c : char;
begin
    if (length(s)=5) and (s[3]='/') then delete(s,3,1)
    else if length(s)<>4 then cc.error('SImg_BayerFlipPattern: InvalidBayer pattern: '+s);

    if flipx then
    begin
        c:=s[1]; s[1]:=s[2]; s[2]:=c;
        c:=s[3]; s[3]:=s[4]; s[4]:=c;
    end;

    if flipy then
    begin
        c:=s[1]; s[1]:=s[3]; s[3]:=c;
        c:=s[2]; s[2]:=s[4]; s[4]:=c;
    end;
    result:=s;
end;


{*************************************************************************}
{******* SImg_BayerInterpolation *****************************************}
{*************************************************************************}
// pass 0: mask known pixels: 0: unknown; 2: known
procedure SImg_BayerInterpolation_pass0(bi:TSImg_BayerInfo; fi:TSImg_FPImage; fj:TSImg_FPImage; mask:pbyte; ya:longint; yz:longint; c:longint);
var x,y,i    : longint;
    d,d1     : psimgfp;
    b        : pbyte;
begin
    fillchar(mask[ya*fj.sl], (yz-ya+1)*fj.sl, 0);
    d1:=fi.pt + ya*fi.sl;
    d:=fj.pt + ya*fj.sl;
    b:=mask + ya*fj.sl;
    if c<0 then
    begin
        for y:=ya to yz do
            for x:=0 to fi.nx-1 do
            begin
                i:=bi.pattern[y and 1, x and 1];
                d[i]:=d1^*bi.cf[i];
                b[i]:=2;
                d1+=1;
                d+=fj.nc;
                b+=fj.nc;
            end;
    end
    else begin
        for y:=ya to yz do
            for x:=0 to fi.nx-1 do
            begin
                if bi.pattern[y and 1, x and 1]=c then
                begin
                    d^:=d1^*bi.cf[c];
                    b^:=2;
                end;
                d1+=1;
                d+=1;
                b+=1;
            end;
    end;
end;


{ pass 1 (diagonal median and all other methods)
    algo < -1 : weight mask
    algo = -1 : q25
    algo =  0 : linear
    algo >  0 : median
}
procedure SImg_BayerInterpolation_pass1(fj:TSImg_FPImage; mask:pbyte; ya:longint; yz:longint; algo:longint);
const weight_undef = 1; { weight interpolation: factor if pixel is undefined. It is 1 because underweighting of undefined pixels should be handled separately. }
var x,y,c      : longint;
    r          : array[0..3] of simgfp;
    s          : simgfp;
    d,d1,d2    : psimgfp;
    b,b1       : pbyte;
begin
    d:=fj.pt + ya*fj.sl;
    b:=mask + ya*fj.sl;
    d1:=d; d2:=d; // makes the compiler happy
    b1:=b;        // makes the compiler happy
    for y:=ya to yz do
    begin
        if y=0 then d1:=d+fj.sl
        else d1:=d-fj.sl;
        if y=fj.ny-1 then d2:=d-fj.sl
        else d2:=d+fj.sl;
        if y=0 then b1:=b+fj.sl
        else b1:=b-fj.sl;

//define[SImg_BayerInterpolation_median1(][,#1);][
        for c:=0 to fj.nc-1 do
        begin
            if (b[0]=0) and (b1[#0]=2) and (b[#0]=0) then
            begin
                r[0]:=d1[#0];
                r[1]:=d1[#1];
                r[2]:=d2[#0];
                r[3]:=d2[#1];
                { median calculated by dropping brightest and darkest value }
                if r[1]<r[0] then
                begin
                    s:=r[0];
                    r[0]:=r[1];
                    r[1]:=s;
                end;    { r0<r1 }
                if r[3]<r[2] then
                begin
                    s:=r[2];
                    r[2]:=r[3];
                    r[3]:=s;
                end;    { r2<r3 }
                if r[0]<r[2] then r[0]:=0
                else r[2]:=0;
                if r[1]>r[3] then r[1]:=0
                else r[3]:=0;
                d^:=0.5*(r[0]+r[1]+r[2]+r[3]);
                b^:=1;
            end;
            d+=1;
            d1+=1;
            d2+=1;
            b+=1;
            b1+=1;
        end;
]

//define[SImg_BayerInterpolation_linear(][,#1);][
        for c:=0 to fj.nc-1 do
        begin
            if b^=0 then
            begin
                if (b1[0]<>0) and (b[#0]<>0) then d^:=0.25*(d1[0]+d2[0]+d[#0]+d[#1])
                else if (b1[0]<>0) then d^:=0.5*(d1[0]+d2[0])
                else if (b[#0]<>0) then d^:=0.5*(d[#0]+d[#1])
                else d^:=0.25*(d1[#0]+d1[#1]+d2[#0]+d2[#1]);
            end;
            d+=1;
            d1+=1;
            d2+=1;
            b+=1;
            b1+=1;
        end;
]

//define[SImg_BayerInterpolation_q25(][,#1);][
        for c:=0 to fj.nc-1 do
        begin
            if b^=0 then
            begin
                if b1[0]=b[#0] then
                begin
                    if b1[0]=0 then
                    begin
                        r[0]:=d1[#0];
                        r[1]:=d1[#1];
                        r[2]:=d2[#0];
                        r[3]:=d2[#1];
                    end
                    else begin
                        r[0]:=d1[0];
                        r[1]:=d2[0];
                        r[2]:=d[#0];
                        r[3]:=d[#1];
                    end;
                    if r[1]<r[0] then
                    begin
                        s:=r[0];
                        r[0]:=r[1];
                        r[1]:=s;
                    end;        { r0<r1 }
                    if r[3]<r[2] then
                    begin
                        s:=r[2];
                        r[2]:=r[3];
                        r[3]:=s;
                    end;        { r2<r3 }
                    if r[1]<r[2] then d^:=0.5*(r[0]+r[1])
                    else if r[3]<r[1] then d^:=0.5*(r[2]+r[3])
                    else d^:=0.5*(r[0]+r[2]);
                end
                else begin
                    if b1[0]=0 then
                    begin
                        r[0]:=d[#1];
                        r[1]:=d[#0];
                    end
                    else begin
                        r[0]:=d2[0];
                        r[1]:=d1[0];
                    end;
                    if r[1]<r[0] then d^:=r[1]
                    else d^:=r[0];
                end;
            end;
            d+=1;
            d1+=1;
            d2+=1;
            b+=1;
            b1+=1;
        end;
]

//define[SImg_BayerInterpolation_weight(][,#1);][
        for c:=0 to fj.nc-1 do
        begin
            if b^=0 then
            begin
                if b1[0]=b[#0] then
                begin
                    if b1[0]=0 then
                    begin
                        r[0]:=d1[#0];
                        r[1]:=d1[#1];
                        r[2]:=d2[#0];
                        r[3]:=d2[#1];
                    end
                    else begin
                        r[0]:=d1[0];
                        r[1]:=d2[0];
                        r[2]:=d[#0];
                        r[3]:=d[#1];
                    end;
                    if r[2]<r[0] then r[0]:=r[2];
                    if r[3]<r[1] then r[1]:=r[3];
                end
                else begin
                    if b1[0]=0 then
                    begin
                        r[0]:=d[#1];
                        r[1]:=d[#0];
                    end
                    else begin
                        r[0]:=d2[0];
                        r[1]:=d1[0];
                    end
                end;
                if r[1]<r[0] then d^:=weight_undef*r[1]
                else d^:=weight_undef*r[0];
            end;
            d+=1;
            d1+=1;
            d2+=1;
            b+=1;
            b1+=1;
        end;
]

        if algo<-1 then
        begin
            SImg_BayerInterpolation_weight(fj.nc,fj.nc);
            for x:=1 to fj.nx-2 do
                SImg_BayerInterpolation_weight(-fj.nc,fj.nc);
            SImg_BayerInterpolation_weight(-fj.nc,-fj.nc);
        end
        else if algo=-1 then
        begin
            SImg_BayerInterpolation_q25(fj.nc,fj.nc);
            for x:=1 to fj.nx-2 do
                SImg_BayerInterpolation_q25(-fj.nc,fj.nc);
            SImg_BayerInterpolation_q25(-fj.nc,-fj.nc);
        end
        else if algo>0 then
        begin
            SImg_BayerInterpolation_median1(fj.nc,fj.nc);
            for x:=1 to fj.nx-2 do
                SImg_BayerInterpolation_median1(-fj.nc,fj.nc);
            SImg_BayerInterpolation_median1(-fj.nc,-fj.nc);
        end
        else begin
            SImg_BayerInterpolation_linear(fj.nc,fj.nc);
            for x:=1 to fj.nx-2 do
                SImg_BayerInterpolation_linear(-fj.nc,fj.nc);
            SImg_BayerInterpolation_linear(-fj.nc,-fj.nc);
        end;
    end;
end;

// pass 2 (orthogonal median)
procedure SImg_BayerInterpolation_pass2(fj:TSImg_FPImage; mask:pbyte; ya:longint; yz:longint);
var x,y,c      : longint;
    r          : array[0..3] of simgfp;
    s          : simgfp;
    d,d1,d2    : psimgfp;
    b          : pbyte;
begin
    d:=fj.pt + ya*fj.sl;
    b:=mask + ya*fj.sl;
    for y:=ya to yz do
    begin
        if y=0 then d1:=d+fj.sl
        else d1:=d-fj.sl;
        if y=fj.ny-1 then d2:=d-fj.sl
        else d2:=d+fj.sl;

//define[SImg_BayerInterpolation_median2(][,#1);][
        for c:=0 to fj.nc-1 do
        begin
            if (b[0]=0) then
            begin
                r[0]:=d1[0];
                r[1]:=d[#0];
                r[2]:=d2[0];
                r[3]:=d[#1];
                if r[1]<r[0] then
                begin
                    s:=r[0];
                    r[0]:=r[1];
                    r[1]:=s;
                end;    { r0<r1 }
                if r[3]<r[2] then
                begin
                    s:=r[2];
                    r[2]:=r[3];
                    r[3]:=s;
                end;    { r2<r3 }
                if r[0]<r[2] then r[0]:=0
                else r[2]:=0;
                if r[1]>r[3] then r[1]:=0
                else r[3]:=0;
                d^:=0.5*(r[0]+r[1]+r[2]+r[3]);
            end;
            d+=1;
            d1+=1;
            d2+=1;
            b+=1;
        end;
]
        SImg_BayerInterpolation_median2(fj.nc,fj.nc);
        for x:=1 to fj.nx-2 do
            SImg_BayerInterpolation_median2(-fj.nc,fj.nc);
        SImg_BayerInterpolation_median2(-fj.nc,-fj.nc);
    end;
end;

threadprocedure SImg_BayerInterpolation_thread(bi:TSImg_BayerInfo; fi:TSImg_FPImage; fj:TSImg_FPImage; mask:pbyte; ya:longint; yz:longint; pass:longint; algo:longint; c:longint);
begin
    if pass=0 then SImg_BayerInterpolation_pass0(bi, fi, fj, mask, ya, yz, c)
    else if pass=1 then SImg_BayerInterpolation_pass1(fj, mask, ya, yz, algo)
    else SImg_BayerInterpolation_pass2(fj, mask, ya, yz)
end;

{ color interpolation
  bi     : Bayer pattern information
  fi     : input
  fj     : output
  algo
    < -1 : weight mask
    = -1 : q25
    =  0 : linear
    >  0 : median
  c
    <  0 : interpolate all colors
    >= 0 : interpolate color c only, where 0 denotes 1st color }
procedure SImg_BayerInterpolation(cc:CSImg_Config; bi:TSImg_BayerInfo; fi:TSImg_FPImage; var fj:TSImg_FPImage; algo,c:longint);
var fk        : TSImg_FPImage;
    i,j,tn,iz : longint;
    mask      : pbyte;
    threads   : PSImg_BayerInterpolation_thread;
begin
    if cc=nil then cc:=SImg_Config_default;
    SImg_IncDebugLevel;
    cc.debug('SImg_BayerInterpolation: fi='+SImg_InfoStr2(fi)+' fj='+SImg_InfoStr2(fj)+' algo='+int2str(algo)+' c='+int2str(c));
    if not SImg_Integrity(fi) then cc.error('SImg_BayerInterpolation: Invalid image: fi');
    if fi.nc<>1 then cc.warning('SImg_BayerInterpolation: Invalid number of colors: fi');
    if not SImg_BayerCheckInfo(bi) then cc.error('SImg_BayerInterpolation: Invalid Bayer pattern');
    if c>=bi.colors then cc.error('SImg_BayerInterpolation: Invalid color');

    if not sImg_OverWriteAlert(fi,fj) then SImg_Free(cc,fj);
    fk.pt:=nil;
    if c<0 then SImg_New(cc, fk,fi.nx,fi.ny,bi.colors)
    else SImg_New(cc, fk,fi.nx,fi.ny,1);
    getmem2(mask,fk.size);

    tn:=cc.threads;
    getmem(threads,tn*sizeof(CSImg_BayerInterpolation_thread));
    for i:=0 to tn-1 do
        threads[i]:=CSImg_BayerInterpolation_thread.create;

    if algo>0 then iz:=2
    else iz:=1;

    for i:=0 to iz do
    begin
        for j:=0 to tn-1 do
            threads[j].setwork(bi, fi, fk, mask, j*fi.ny div tn,((j+1)*fi.ny div tn)-1, i, algo, c);
        for j:=0 to tn-1 do
            threads[j].wait;
    end;

    for i:=0 to tn-1 do
        threads[i].destroy;
    freemem(threads);

    freemem2(mask,fk.size);
    SImg_Move(cc, fk,fj);

    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_BayerDownScale *********************************************}
{*************************************************************************}
procedure SImg_BayerDownScale(cc:CSImg_Config; bi:TSImg_BayerInfo; fi:TSImg_FPImage; var fj:TSImg_FPImage);
var fk      : TSImg_FPImage;
    x,y,i   : longint;
    d,e     : ^simgfp;
    cn      : array[0..3] of longint;
    cf      : array[0..3] of double;
begin
    if cc=nil then cc:=SImg_Config_default;
    SImg_IncDebugLevel;
    cc.debug('SImg_BayerDownScale: fi='+SImg_InfoStr2(fi)+' fj='+SImg_InfoStr2(fj));
    if not SImg_Integrity(fi) then cc.error('SImg_BayerDownScale: Invalid image: fi');
    if fi.nc<>1 then cc.warning('SImg_BayerDownScale: Invalid number of colors: fi');
    if not SImg_BayerCheckInfo(bi) then cc.error('SImg_BayerDownScale: Invalid Bayer pattern');

    fillchar(cn, sizeof(cn), 0);
    for y:=0 to 1 do
        for x:=0 to 1 do
            cn[bi.pattern[y,x]]+=1;
    for y:=0 to 3 do
    begin
        cf[y]:=bi.cf[y];
        if cn[y]>1 then cf[y]/=cn[y];
    end;

    if not sImg_OverWriteAlert(fi,fj) then SImg_Free(cc,fj);
    fk.pt:=nil; SImg_New(cc, fk,fi.nx shr 1,fi.ny shr 1,bi.colors);
    for y:=0 to fk.ny-1 do
    begin
        d:=fi.pt + y*2*fi.sl;
        e:=fk.pt + y*fk.sl;
        for x:=0 to fk.nx-1 do
        begin
            i:=bi.pattern[0,0];   e[i]+=d[0]*cf[i];
            i:=bi.pattern[0,1];   e[i]+=d[1]*cf[i];
            i:=bi.pattern[1,0];   e[i]+=d[fi.sl]*cf[i];
            i:=bi.pattern[1,1];   e[i]+=d[fi.sl+1]*cf[i];
            e+=fk.nc;  d+=2;
        end;
    end;
    SImg_Move(cc, fk,fj);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_BayerFlat **************************************************}
{*************************************************************************}
procedure SImg_BayerFlat(cc:CSImg_Config; fi:TSImg_FPImage; var bi:TSImg_BayerInfo);
const quant = 0.99; { reference quantile for 1.0 }
var x,y,i : longint;
    d       : ^simgfp;
    g       : array[0..3] of double;
    r,s     : simgfp;
    pat     : TSImg_BayerPattern;
    fj      : TSImg_FPImage;

begin
    if cc=nil then cc:=SImg_Config_default;
    SImg_IncDebugLevel;
    cc.debug('SImg_BayerFlat: fi='+SImg_InfoStr2(fi));
    if not SImg_Integrity(fi) then cc.error('SImg_BayerFlat: Invalid image: fi');
    if fi.nc<>1 then cc.error('SImg_BayerFlat: Invalid number of colors: fi');
    if not SImg_BayerCheckInfo(bi) then cc.error('SImg_BayerFlat: Invalid Bayer pattern');

    fj.pt:=nil;
    SImg_BayerDownScale(cc, bi, fi, fj);
    for i:=0 to 3 do
        g[i]:=1;
    for i:=0 to fj.nc-1 do
    begin
        s:=SImg_GetQuantile(cc,fj,i,quant);
        if s>1e-20 then g[i]:=1/s
        else cc.warning('SImg_BayerFlat: Values for color '+int2str(i)+' to small');
    end;
    SImg_Free(cc,fj);
    pat:=bi.pattern;
    d:=fi.pt;
    s:=1+2*(1-quant);  // maximum
    for y:=0 to fi.ny-1 do
        for x:=0 to fi.nx-1 do
        begin
            r:=d^*g[pat[y and 1, x and 1]];
            if r>s then r:=s;
            d^:=r; d+=1;
        end;
    SImg_DecDebugLevel;
end;

{*************************************************************************}
{******* SImg_BayerScale *************************************************}
{*************************************************************************}
procedure SImg_BayerScale(cc:CSImg_Config; fi:TSImg_FPImage; var bi:TSImg_BayerInfo);
var x,y,i   : longint;
    d       : ^simgfp;
    r       : array[0..4] of double;
    pat     : TSImg_BayerPattern;

begin
    if cc=nil then cc:=SImg_Config_default;
    SImg_IncDebugLevel;
    cc.debug('SImg_BayerScale: fi='+SImg_InfoStr2(fi));
    if not SImg_Integrity(fi) then cc.error('SImg_BayerScale: Invalid image: fi');
    if fi.nc<>1 then cc.error('SImg_BayerScale: Invalid number of colors: fi');
    if not SImg_BayerCheckInfo(bi) then cc.error('SImg_BayerScale: Invalid Bayer pattern');

    pat:=bi.pattern;
    for i:=0 to 3 do
    begin
        r[i]:=bi.cf[i];
        bi.cf[i]:=1;
    end;

    d:=fi.pt;
    for y:=0 to fi.ny-1 do
        for x:=0 to fi.nx-1 do
        begin
            d^*=r[pat[y and 1, x and 1]];
            d+=1;
        end;

    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_ToBayer ****************************************************}
{*************************************************************************}
procedure SImg_ToBayer(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; bi:TSImg_BayerInfo);
var fk       : TSImg_FPImage;
    x,y     : longint;
    d,e     : ^simgfp;

begin
    if cc=nil then cc:=SImg_Config_default;
    SImg_IncDebugLevel;
    cc.debug('SImg_ToBayer: fi='+SImg_InfoStr2(fi) + ' fj='+SImg_InfoStr2(fj));
    if not SImg_Integrity(fi) then cc.error('SImg_ToBayer: Invalid image: fi');
    if not SImg_BayerCheckInfo(bi) then cc.error('SImg_ToBayer: Invalid Bayer pattern');
    if fi.nc<bi.colors then cc.error('SImg_ToBayer: Invalid number of colors: fi');

    if not SImg_OverWriteAlert(fi,fj) then SImg_Free(cc,fj);
    fk.pt:=nil; SImg_New(cc, fk, fi.nx, fi.ny, 1);

    d:=fi.pt; e:=fk.pt;
    for y:=0 to fi.ny-1 do
        for x:=0 to fi.nx-1 do
        begin
            e^:=d[bi.pattern[y and 1, x and 1]];
            d+=fi.nc; e+=1;
        end;

    SImg_Move(cc, fk,fj);
    SImg_DecDebugLevel;
end;
