{  simg -- image processing library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

// define[debug_deconv]

{ Masked Richardson-Lucy o_(k+1) = o_k*( ((i*u)/(o_k \ast s)) \ast s~ ) / (u \ast s~) where u is the carrier of o
    fi          : input
    fo          : output
    rm          : mask, bit0 or bit 2: constraint mask, bit 1 saturated
    cfi         : correct fo
}
(*
procedure SImg_DeconvMasked(cc:CSImg_Config; fi,fo:TSImg_FPImage; rm:TSImg_RDImage;  rad:double; it:longint);
const rl_reg = 1e-2;    // RL regularization
      u_reg  = 1e-10;   // saturation mask regularization
var fj,fm  : TSImg_FPImage;
    i,j : longint;
    r,s : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_DeconvMasked: Invalid image: fi');
    if not SImg_Recover(cc,fo) then cc.error('SImg_DeconvMasked: Invalid image: fo');
    if fo.size<>fi.size then cc.error('SImg_DeconvMasked: Invalid dimensions: fo');
    if not SImg_Recover(cc,rm) then cc.error('SImg_DeconvMasked: Invalid image: rm');
    if rm.fmt<>SImg_RawByte then cc.error('SImg_DeconvMasked: Invalid data format: rm');
    if rm.size<>fi.size then cc.error('SImg_DeconvMasked: Invalid dimensions: rm');

    { mask factor 1/(u \ast s~) }
    fm.pt:=nil;
    SImg_New(cc, fm, fi.nx,fi.ny,1, false);
    s:=0;
    for i:=0 to fi.size-1 do
    begin
        j:=rm.ptb[i];
        if j and 5=0 then fo.pt[i]:=0;  // constraint mask
        if j and 2=0 then fm.pt[i]:=1   // saturation mask
        else fm.pt[i]:=u_reg;
        s+=fi.pt[i];
    end;
    SImg_Filter(cc, fm,fm,rad);
    for i:=0 to fm.size-1 do
        fm.pt[i]:=1/fm.pt[i];
    s*=rl_reg/fi.size;

    { RL iterations }
    fj.pt:=nil;
    for i:=1 to it do
    begin
        SImg_Filter(cc, fo,fj,rad);
        for j:=0 to fi.size-1 do
        begin
            r:=fi.pt[j];
            if r<0 then r:=0;
            r:=(r+s)/(fj.pt[j]+s);
            if rm.ptb[j] and 2<>0 then r*=u_reg;
            fj.pt[j]:=r;
        end;
        SImg_Filter(cc, fj,fj,rad);
        for j:=0 to fo.size-1 do
            fo.pt[j]*=fj.pt[j]*fm.pt[j];
    end;
    SImg_free(cc,fj);
    SImg_free(cc,fm);
end;
*)


{*************************************************************************}
{******* SImg_PSFWisdom **************************************************}
{*************************************************************************}
{ Extract PSF wisdom from samples
  fc   : PSF wisdom buffer. Only the upper limit2 is saved <because lower limit>(x,y)=<upper limit)(nx-1-x,ny-1-y)
  fo   : object (ofwmh)
  fi   : image (ifwmh)
  ifwhm : input FWHM
  ofwhm : output FWHM
  gsize : FG/BG approx. size
}
procedure SImg_PSFWisdom_extract(cc:CSImg_Config; fc, fo,fi:TSImg_FPImage; ifwhm,ofwhm,gsize:double);
var fj,fmini,fmaxi,fminj,fmaxj  : TSIMG_FPimage;
    i,j,x,y                     : longint;
    regi,regj,ri,rj,r           : double;
begin
    { filters }
    fj.pt:=nil;   SImg_Filter(cc, fi,fj, ifwhm*FWHM_TO_RAD);
    fmini.pt:=nil; SImg_FBGFilter(cc, fi,fmini, gsize, true);
    fmaxi.pt:=nil; SImg_FBGFilter(cc, fi,fmaxi, gsize, false);
    fminj.pt:=nil; SImg_FBGFilter(cc, fj,fminj, gsize*sqrt(2), true);
    fmaxj.pt:=nil; SImg_FBGFilter(cc, fj,fmaxj, gsize*sqrt(2), false);
    { fmin, fmax regularization }
    regi:=1e-100;
    regj:=1e-100;
    for i:=0 to fmaxi.size-1 do
    begin
        if fmaxi.pt[i]>regi then regi:=fmaxi.pt[i];
        if fmaxj.pt[i]>regj then regj:=fmaxj.pt[i];
    end;
    regi*=1e-4;
    regj*=1e-4;
    { extract limits }
    for i:=0 to fo.size-1 do
        if (fmaxi.pt[i]-fmini.pt[i]>regi) and (fmaxj.pt[i]-fminj.pt[i]>regj) then
        begin
            r:=fj.pt[i];
            ri:=1/(fmaxi.pt[i]-fmini.pt[i]);
            rj:=1/(fmaxj.pt[i]-fminj.pt[i]);
            x:=round((r-fmini.pt[i])*ri*(fc.nx-1));
            if x<0 then x:=0;
            if x>=fc.nx then x:=fc.nx-1;
            y:=round((r-fminj.pt[i])*rj*(fc.ny-1));
            if y<0 then y:=0;
            if y>=fc.ny then y:=fc.ny-1;
            r:=fo.pt[i]-r;
            j:=(y*fc.nx+x)*2;                   { upper limit }
            if r<0 then j:=fc.size-2-j;         { lower limit }
            ri*=abs(r);
            rj*=abs(r);
            if ri>fc.pt[j  ] then fc.pt[j  ]:=ri;
            if rj>fc.pt[j+1] then fc.pt[j+1]:=rj;
        end;
    SImg_Free(cc,fj);
    SImg_Free(cc,fmini);
    SImg_Free(cc,fmaxi);
    SImg_Free(cc,fminj);
    SImg_Free(cc,fmaxj);
end;


{ Calculate a sample pattern
  fs    : PSF
  fi2   : sample object convoluted with PSF
  ns    : disable stars / points
}
procedure SImg_PSFWisdom_sample(cc:CSImg_Config; fs:TSImg_FPImage; var fi2:TSImg_FPImage; ns:boolean);
    function samplef(x,y,n:longint):longint;
    begin
        result:=1;
        if (x<0) or (x>n*8) or (y<0) or (y>n*8) then result:=0
        else if (y<n) and (x>4*n) and (x<6*n) then
        begin
            if sqr(x-5*n)+sqr(y)<sqr(n) then result:=0;
        end
        else if (y<2*n) and (x>6*n) then
        begin
            if sqr(x-6*n)+sqr(y-2*n)>sqr(2*n) then result:=0;
        end
        else if (y>4*n) and (x>4*n) then
        begin
            if sqr(x-4*n)+sqr(y-4*n)>sqr(4*n) then result:=0;
        end;
    end;

    // x,y: 0..n*20-1
    // n: >=2
    function sample(x,y,n:longint; ns:boolean):longint;
    var a00,a01,a10,a11  : longint;
    begin
        { dots }
        if x>=n*12 then
        begin
            a00:=(n*(y div (n*4))) div 2;       // radius
            if ns then a00+=n div 2+1;
            if x>=n*16 then a01:=n div 2        // y stretch
            else a01:=0;
            a10:=round(sqr(n)/((a00+a01+1)*(a00+1)));   // return value
            if a10<1 then a10:=1;
            x:=(x mod (n*4)) - 2*n;
            y:=(y mod (n*4)) - 2*n;
            if y<0 then
            begin
                y+=a01;
                if y>0 then y:=0;
            end;
            if sqr(x)+sqr(y)<=sqr(a00) then result:=a10
            else result:=0;
        end
        else if y<=n*10 then result:=samplef(x,y,n)     // filled figure
        else begin
            y-=n*12-2;
            a00:=samplef(x-1,y-1,n);
            a01:=samplef(x-1,y  ,n);
            a10:=samplef(x  ,y-1,n);
            a11:=samplef(x  ,y  ,n);
            result:=n*(abs(a00-a01) + abs(a10-a11) + abs(a00-a10) + abs(a01-a11)) div 2;
        end;
    end;

var fi          : TSImg_FPImage;
    n,x,y,u,v,b : longint;
    d,e         : ^simgfp;
begin
    n:=round(fs.nx/4);
    if n<2 then n:=2;
    SImg_New(cc, fi2, n*20+fs.nx, n*20+fs.ny, 1); fi:=fi2;
//ifdef[debug_deconv]
    for y:=0 to n*20-1 do
        for x:=0 to n*20-1 do
            fi.pt[y*fi.sl + x]:=sample(x,y,n,ns);
    SImg_WriteImageFile(cc,'psfwisdom_sample-n='+int2str(n)+'.fits',fi);
    SImg_Clear(cc,fi);
//endif[debug_deconv]
    for y:=0 to n*20-1 do
        for x:=0 to n*20-1 do
        begin
            b:=sample(x,y,n,ns);
            if b>0 then
                for u:=0 to fs.ny-1 do
                begin
                    d:=fi.pt + (y+u)*fi.sl + x;
                    e:=fs.pt + u*fs.sl;
                    for v:=0 to fs.nx-1 do
                        d[v]+=e[v]*b;
                end;
        end;
end;


{ Calculate constraints map (PSF wisdom) from sample pattern
  psfw  : constraint map
  ifwhm : input FWHM
  ofwhm : output FWHM
  gsize : FG/BG approx. size
  ns    : disable stars / points
}
procedure SImg_PSFWisdom(cc:CSImg_Config; var psfw:TSImg_RDImage; ifwhm,ofwhm,gsize:double; ns:boolean);
const x_offset   = 0.2;
      y_offset   = 0.1;
      size_fg    = 3;
      size       = size_fg*15;
var fso,fsi,fc,fi,fo    : TSImg_FPImage;
    x,y,i,n             : longint;
    a0,a1,b0,b1,cx,cy,r,s : double;

begin
    { calculate PSF's }
    n:=round(ifwhm*6);
    if n<ofwhm*6 then n:=round(ofwhm*6);
    cx:=round(0.5*n)+x_offset;
    cy:=round(0.5*n)+y_offset;
    fso.pt:=nil; SImg_New(cc, fso, n+1,n+1,1);
    fsi.pt:=nil; SImg_New(cc, fsi, n+1,n+1,1);
    a0:=-1/sqr(ofwhm*FWHM_TO_RAD);
    a1:=-1/sqr(ifwhm*FWHM_TO_RAD);
    b0:=0; b1:=0;
    for y:=0 to n do
        for x:=0 to n do
        begin
            r:=sqr(y-cy)+sqr(x-cx);
            i:=y*(n+1)+x;
            s:=exp(r*a0);
            fso.pt[i]:=s;
            b0+=s;
            s:=exp(r*a1);
            fsi.pt[i]:=s;
            b1+=s;
        end;
    r:=b1/b0;
    for i:=0 to fso.size-1 do
        fso.pt[i]*=r;

    { wisdom extraction from sample }
    fc.pt:=nil; SImg_New(cc, fc, size,size,2 );
    fi.pt:=nil; fo.pt:=nil;
    SImg_PSFWisdom_sample(cc, fso, fo, ns);
    SImg_PSFWisdom_sample(cc, fsi, fi, ns);
    SImg_Free(cc,fso);
    SImg_Free(cc,fsi);
//ifdef[debug_deconv]
    SImg_WriteImageFile(cc,'psfwisdom_fo.fits',fo);
    SImg_WriteImageFile(cc,'psfwisdom_fi.fits',fi);
//endif
    SImg_PSFWisdom_extract(cc, fc, fo,fi, ifwhm,ofwhm,gsize);

    SImg_Filter(cc, fo,fo, ifwhm/6);
    SImg_Filter(cc, fi,fi, ifwhm/6);
    SImg_PSFWisdom_extract(cc, fc, fo,fi, ifwhm,ofwhm,gsize);

    SImg_Filter(cc, fo,fo, ifwhm/4);
    SImg_Filter(cc, fi,fi, ifwhm/4);
    SImg_PSFWisdom_extract(cc, fc, fo,fi, ifwhm,ofwhm,gsize);

    SImg_Filter(cc, fo,fo, ifwhm*0.5);
    SImg_Filter(cc, fi,fi, ifwhm*0.5);
    SImg_PSFWisdom_extract(cc, fc, fo,fi, ifwhm,ofwhm,gsize);

//ifdef[debug_deconv]
    SImg_WriteImageFile(cc,'psfwisdom_fc.fits',fc);
//endif
    SImg_Free(cc,fo);
    SImg_Free(cc,fi);

    { convert to bytes }
    SImg_New(cc, psfw, SImg_RawByte, size, size, 2, false);
    SImg_FBGFilter(cc, fc,fc, size_fg, false);
    r:=1e-100;
    for i:=0 to fc.size-1 do
        if fc.pt[i]>r then r:=fc.pt[i];
    psfw.cf:=r/255;
    r:=255/r;
    for i:=0 to fc.size-1 do
        psfw.ptb[i]:=round(fc.pt[i]*r);
//ifdef[debug_deconv]
    SImg_WriteImageFile(cc,'psfwisdom_psfw.fits',fc);
//endif

    SImg_Free(cc,fc);
end;


{*************************************************************************}
{******* SImg_FFTPSF *****************************************************}
{*************************************************************************}
{ Calculate Fourier transform of Gauss shaped PSF
    rad   : PSF radius
    ti    : output image
    nx,ny : dimensions of the image to (de)convolute
}
procedure SImg_FFTPSF(cc:CSImg_Config; rad:double; var ti:TSImg_FFTImage; nx,ny:longint);
var rr,r,s      : double;
    d           : ^fftfp;
    x,y,n,i     : longint;
begin
    rad:=abs(rad);
    if rad<1e-6 then rad:=1e-6;
    rr:=-sqr(1/rad);

    nx+=round(rad*8+1);
    ny+=round(rad*8+1);

    n:=round(rad*4+1);
    SImg_New_findbase(cc, ti, nx,ny);

    { normalized PSF }
    s:=0;
    for y:=-n to n do
    begin
        i:=y;
        if y<0 then i+=ti.ny;
        d:=ti.pt + i*ti.nxc*2;
        for x:=-n to n do
        begin
            i:=x;
            if i<0 then i+=ti.nxr;
            r:=exp((sqr(x)+sqr(y))*rr);
            s+=r;
            d[i]:=r;
        end;
    end;
    s:=1/s;
    for y:=-n to n do
    begin
        i:=y;
        if y<0 then i+=ti.ny;
        d:=ti.pt + i*ti.nxc*2;
        for x:=-n to n do
        begin
            i:=x;
            if i<0 then i+=ti.nxr;
            d[i]*=s;
        end;
    end;
//    SImg_FFTWrite(cc,'fftpsf-psf.fits',ti);

    { FFT transform }
    SImg_FFT_r2c(cc,ti);
end;


{*************************************************************************}
{******* SImg_FFTWiener **************************************************}
{*************************************************************************}
{ Wiener filter in frequency space
    snr   : SNR
    ti    : input Fourier-transformed image
    tj    : output Fourier-transformed image (can be equal to ti or ts)
    ts    : Fourier-transformed PSF
}
procedure SImg_FFTWiener(cc:CSImg_Config; snr:double; ti,tj,ts:TSImg_FFTImage);
var r,sr,sc,ir,ic,sn    : double;
    i                   : longint;
begin
    snr:=abs(snr);
    if snr<0.1 then snr:=0.1;
    if (ts.nxc<>ti.nxc) or (ts.ny<>ti.ny) then cc.error('SImg_FFTWiener: invalid dimensions: ti');
    if (ts.nxc<>tj.nxc) or (ts.ny<>tj.ny) then cc.error('SImg_FFTWiener: invalid dimensions: tj');

    sn:=(sqr(ts.pt[0])+simg_mina)/sqr(snr);     { estimated spectral power density of noise }
    for i:=0 to ti.nxc*ti.ny-1 do
    begin
        sr:=ts.pt[i*2];
        sc:=ts.pt[i*2+1];
        ir:=ti.pt[i*2];
        ic:=ti.pt[i*2+1];
        r:=1/(sqr(sr)+sqr(sc)+sn);
        tj.pt[i*2  ]:=(sr*ir+sc*ic)*r;
        tj.pt[i*2+1]:=(sr*ic-sc*ir)*r;
    end;
end;

{ Calculate inverse filter using Wiener regularization with constant spectrum
    snr   : SNR
    ts    : Fourier-transformed PSF
    tj    : output Fourier-transformed inverse filter (can be equal to ts)
}
procedure SImg_FFTWiener(cc:CSImg_Config; snr:double; ts,tj:TSImg_FFTImage);
var r,sr,sc,sn  : double;
    i           : longint;
begin
    snr:=abs(snr);
    if snr<0.1 then snr:=0.1;
    if (ts.nxc<>tj.nxc) or (ts.ny<>tj.ny) then cc.error('SImg_FFTWiener: invalid dimensions: tj');

    sn:=(sqr(ts.pt[0])+simg_mina)/sqr(snr);     { estimated spectral power density of noise }
    for i:=0 to ts.nxc*ts.ny-1 do
    begin
        sr:=ts.pt[i*2];
        sc:=ts.pt[i*2+1];
        r:=1/(sqr(sr)+sqr(sc)+sn);
        tj.pt[i*2  ]:= sr*r;
        tj.pt[i*2+1]:=-sc*r;
    end;
end;


{*************************************************************************}
{******* SImg_DeconvWiener ***********************************************}
{*************************************************************************}
{ Wiener deconvolution with constraints
    fi          : input
    fo          : initial solution (optional) + output
    fmin,fmax   : constraints. Ignored if fmin.pt=nil. 0 will be uesed instead
    rad         : deconvolution radius
    snr         : Regularization for Wiener filter as SNR.
    cit         : constraint iterations
}
procedure SImg_DeconvWiener(cc:CSImg_Config; fi:TSImg_FPImage; var fo:TSImg_FPImage; fmin,fmax:TSImg_FPImage; rad,snr:double; cit:longint);
var fk          : TSImg_FPImage;
    ts,ti       : TSImg_FFTImage;
    r,s,t,u     : double;
    ci,x,y      : longint;
    d           : ^fftfp;
    e,f,g,h     : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_DeconvWiener: Invalid image: fi');
    if fi.nc<>1 then cc.error('SImg_DeconvWiener: Invalid dimensions: fi');
    if fo.pt<>nil then
    begin
        if not SImg_Recover(cc,fo) then cc.error('SImg_DeconvWiener: Invalid image: fo');
        if fo.size<>fi.size then cc.error('SImg_DeconvWiener: Invalid dimensions: fo');
    end;
    if fmin.pt<>nil then
    begin
        if not SImg_Recover(cc,fmin) then cc.error('SImg_DeconvWiener: Invalid image: fmin');
        if fmin.size<>fi.size then cc.error('SImg_DeconvWiener: Invalid dimensions: fmin');
        if not SImg_Recover(cc,fmax) then cc.error('SImg_DeconvWiener: Invalid image: fmax');
        if fmax.size<>fi.size then cc.error('SImg_DeconvWiener: Invalid dimensions: fmax');
    end;

    ts.pt:=nil; ti.pt:=nil; fk.pt:=nil;
    SImg_FFTPSF(cc, rad, ts, fi.nx,fi.ny);

    { residual }
    if fo.pt<>nil then
    begin
        SImg_Filter(cc, fo,fk, rad);
        SImg_New(cc, ti, ts.nxr,ts.ny);
        for y:=0 to fi.ny-1 do
        begin
            d:=ti.pt + y*ti.nxc*2;
            e:=fi.pt + y*fi.sl;
            f:=fk.pt + y*fk.sl;
            for x:=0 to fi.nx-1 do
            begin
                r:=e[x]-f[x];
                d[x]:=r;
            end;
        end;
        SImg_Free(cc,fk);
    end
    else begin
        SImg_New(cc, ti, ts.nxr,ts.ny);
        SImg_Copy(cc, fi,ti,0);
    end;

    { wiener filter }
//  if i=0 then SImg_FFTWrite(cc,'deconvdenoise-res0.fits',ti);
    SImg_FFT_r2c(cc,ti);
    SImg_FFTWiener(cc, snr, ti,ti,ts);
    SImg_FFT_c2r(cc,ti);
//  if i=0 then SImg_FFTWrite(cc,'deconvdenoise-wiener0.fits',ti);
    snr*=sqrt(2);
    { constraints }
    if cit>0 then SImg_New(cc, fk, fi.nx,fi.ny,1);
    h:=fk.pt;
    if fo.pt=nil then SImg_New(cc, fo, fi.nx,fi.ny,1);
    for y:=0 to fi.ny-1 do
    begin
        d:=ti.pt + y*ti.nxc*2;
        e:=fo.pt + y*fo.sl;
        f:=fmin.pt + y*fmin.sl;
        g:=fmax.pt + y*fmax.sl;
        for x:=0 to fi.nx-1 do
        begin
            r:=e[x]+d[x]; t:=r;
            if fmin.pt<>nil then
            begin
                s:=f[x]; if r<s then r:=s;
                s:=g[x]; if r>s then r:=s;
            end else
            begin
                if r<0 then r:=0;
            end;
            e[x]:=r;
            if h<>nil then
            begin
                h^:=t-r;
                h+=1;
            end;
        end;
    end;
    SImg_Free(cc,ti);

    { constraint iterations }
    for ci:=1 to cit do
    begin
        SImg_Filter(cc, fk,fk, rad);
        u:=0;
        for y:=0 to fi.ny-1 do
        begin
            e:=fo.pt   + y*fo.sl;
            f:=fmin.pt + y*fmin.sl;
            g:=fmax.pt + y*fmax.sl;
            h:=fk.pt   + y*fk.sl;
            for x:=0 to fi.nx-1 do
            begin
                r:=e[x]+h[x]; t:=r;
                if fmin.pt<>nil then
                begin
                    s:=f[x]; if r<s then r:=s;
                    s:=g[x]; if r>s then r:=s;
                end else
                begin
                    if r<0 then r:=0;
                end;
                e[x]:=r;
                t-=r;
                h[x]:=t;
                u+=sqr(t);
            end;
        end;
        cc.info('Constraint iteration '+int2str(ci)+': Residual='+real2str(sqrt(u/fi.size)));
    end;
//  SImg_WriteImageFile(cc,'devonv_wiener-fk-'+int2str(wi)+'.fits',fk);
    SImg_Free(cc,fk);
    SImg_Free(cc,ts);
end;


{*************************************************************************}
{********* SImg_DeconvRL *************************************************}
{*************************************************************************}
{ Richardson Lucy deconvolution
  fi      : input
  fo      : initial solution and output
  rad     : star radius
  it      : iterations
}
procedure SImg_DeconvRL(cc:CSImg_Config; fi,fo:TSImg_FPImage; rad:double; it:longint);
var fj  : TSImg_FPImage;
    i,j : longint;
    r,s : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_DeconvMasked: Invalid image: fi');
    if not SImg_Recover(cc,fo) then cc.error('SImg_DeconvMasked: Invalid image: fo');
    if fo.size<>fi.size then cc.error('SImg_DeconvMasked: Invalid dimensions: fo');

    fj.pt:=nil;
    for i:=1 to it do
    begin
        SImg_Filter(cc, fo,fj,rad);
        for j:=0 to fi.size-1 do
        begin
            r:=fj.pt[j];
            s:=fi.pt[j];
            if s<simg_mina then fj.pt[j]:=0
            else if (r<simg_mina) or (r<1e-2*s) then fj.pt[j]:=100
            else fj.pt[j]:=s/r;
        end;
        SImg_Filter(cc, fj,fj,rad);
        for j:=0 to fo.size-1 do
            fo.pt[j]*=fj.pt[j];
    end;
    SImg_free(cc,fj);
end;


{*************************************************************************}
{******* SImg_DeconvDenoise **********************************************}
{*************************************************************************}
{ Calculate and apply constraints
   ti,tn : thread index/number
   fi    : image
   fmini : BG filtered fi, lower constraint output
   fmaxi : FG filtered fi - fmini, upper constraint output
   fj    : ifwhm-filtered fi, output
   fminj : BG filtered fj
   fmaxj : FG filtered fj - fminj
   psfw  : PSF wisdom
   dn    : denoising strength
   dni   : mini/maxi reduction
   dnj   : minj/maxj reduction
   ddn0  : brightness limit for 100% denoising
   ddn1  : brightness limit for 0% denoising
   bdn   : denoising at boundaries relative to image dimensions
}
threadprocedure SImg_DeconvDenoise_Constraints_thread(ti:longint; tn:longint; fi:TSImg_FPImage; fmini:TSImg_FPImage; fmaxi:TSImg_FPImage; fj:TSImg_FPImage; fminj:TSImg_FPImage; fmaxj:TSImg_FPImage; psfw:TSImg_RDImage; dn:double; dni:double; dnj:double; ddn0:double; ddn1:double; bdn:double);
var i,j,x,y,yi,ya,yz,xi : longint;
    r,s,t,ri,rj,lu,ll   : double;
    dn2,dny,rbx,rby,rddn: double;
begin
    ya:=(fi.ny*ti) div tn;
    yz:=(fi.ny*(ti+1)) div tn - 1;
    dni*=dn;
    dnj*=dn;
    dn:=sqr(1-dn);
    if bdn>0.5 then bdn:=0.5;
    if bdn<1e-12 then bdn:=1e-12;
    bdn:=1/(4*bdn);
    rbx:=2*bdn/(fi.nx+1);
    rby:=2*bdn/(fi.ny+1);
    rddn:=ddn1-ddn0;
    if rddn<simg_mina then rddn:=0
    else rddn:=1/rddn;
    i:=ya*fi.sl;
    for yi:=ya to yz do
    begin
        dny:=abs((yi+1)*rby-bdn)-bdn+1;
        if dny<0 then dny:=1
        else dny:=1-sqr(dny);
        for xi:=0 to fi.nx-1 do
        begin
            r:=fj.pt[i];
            ri:=fmaxi.pt[i]{-fmini.pt[i]};
            rj:=fmaxj.pt[i]{-fminj.pt[i]};
            if (r>ddn0) and (ri>simg_mina) and (rj>simg_mina) then
            begin
                { extra denoising: achieved by reduction of difference between constraints and fj }
                dn2:=abs((xi+1)*rbx-bdn)-bdn+1;
                if dn2<0 then dn2:=dny
                else dn2:=dny*(1-sqr(dn2));
                if r<ddn1 then dn2*=(r-ddn0)*rddn;
                { PSF wisdom coordinates }
                s:=fmini.pt[i];
                x:=round((r-s)/ri*(psfw.nx-1));
                if x<0 then x:=0;
                if x>=psfw.nx then x:=psfw.nx-1;
                y:=round((r-fminj.pt[i])/rj*(psfw.ny-1));
                if y<0 then y:=0;
                if y>=psfw.ny then y:=psfw.ny-1;
                { minimum limits }
                lu:=(ri+s-r)*dn;
                ll:=(r-s)*dn;
                { constant area noise reduction }
                ri:=(ri-dni)*psfw.cf;
                if ri<0 then ri:=0;
                rj:=(rj-dnj)*psfw.cf;
                if rj<0 then rj:=0;
                { upper limit }
                j:=(y*psfw.nx+x)*2;
                s:=ri*psfw.ptb[j  ];
                t:=rj*psfw.ptb[j+1];
                if s<t then t:=s;
                if t>lu then lu:=t;
                lu:=r+lu*dn2;
                { lower limit }
                j:=psfw.size-2-j;
                s:=ri*psfw.ptb[j  ];
                t:=rj*psfw.ptb[j+1];
                if s<t then t:=s;
                if t>ll then ll:=t;
                ll:=r-ll*dn2;
                { output }
                r:=fi.pt[i];
                if r<ll then r:=ll;
                if r>lu then r:=lu;
                fmini.pt[i]:=ll;
                fmaxi.pt[i]:=lu;
            end
            else begin
                fmini.pt[i]:=r;
                fmaxi.pt[i]:=r;
            end;
            fj.pt[i]:=r;
            i+=1;
        end;
    end;
end;

{******* SImg_DeconvDenoise **********************************************}
{ Deconvolution and denoising
    fi     : input image (1 color image)
    fo     : output image
    ifwhm  : input FWHM
    ofwhm  : output FWHM
    dn     : denoising strength (0..1)
    ddn    : Additional denoising of regions darker than that quantile.
    bdn    : Additional denoising at boundaries. Relative to image dimension. 0..0.5
    bg     : extra lower constraints. Recommended: 0..1. <0.1 means disable
    ns     : no star mode
    snr    : SNR parameter for wiener deconvolution
}
procedure SImg_DeconvDenoise(cc:CSImg_Config; fi:TSImg_FPImage; var fo:TSImg_FPImage; ifwhm,ofwhm, dn,ddn,bdn, bg:double; ns:boolean; snr:double);
const fwhm_to_gsize = 0.5;
      cit           = 5;        { constraint iterations }
var fmini,fmaxi         : TSImg_FPImage;
    fj,fminj,fmaxj      : TSImg_FPImage;
    psfw                : TSImg_RDImage;
    i,tn                : longint;
    gsize,dni,dnj,r     : double;
    ddn0,ddn1           : double;
    dc                  : boolean;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_DeconvDenoise: fi='+SImg_InfoStr2(fi)+' fo='+SImg_InfoStr2(fo)+' ifwhm='+real2str(ifwhm)+' ofwhm='+real2str(ofwhm)+' dn='+real2str(dn)+' ddn='+real2str(ddn)+
        ' bdn='+real2str(bdn)+' bg='+real2str(bg)+' ns='+boolean2str(ns)+' snr='+real2str(snr));
    if not SImg_Recover(cc,fi) then cc.error('SImg_Denoise_getFilter: Invalid image: fi');
    if fi.nc<>1 then cc.error('SImg_Denoise_getFilter: One color image expected: fi');
    if not SImg_OverwriteAlert(fi,fo) then SImg_Free(cc,fo);
    snr:=abs(snr);
    if snr<1 then snr:=1;
    if ofwhm<1.1 then ofwhm:=1.1;
    if ifwhm<1.1 then ifwhm:=1.1;
    gsize:=ifwhm*fwhm_to_gsize;
    dc:=sqr(ifwhm)-sqr(ofwhm)>0.25;

    { contraints }
    psfw.pt:=nil;  SImg_PSFWisdom(cc, psfw, ifwhm,ofwhm,gsize, ns);
    fmini.pt:=nil; SImg_FBGFilter(cc, fi,fmini, gsize, true);
    fmaxi.pt:=nil; SImg_FBGFilter(cc, fi,fmaxi, gsize, false);
    fj.pt:=nil;    SImg_Filter(cc, fi,fj, ifwhm*FWHM_TO_RAD);
    fminj.pt:=nil; SImg_FBGFilter(cc, fj,fminj, gsize*sqrt(2), true);
    fmaxj.pt:=nil; SImg_FBGFilter(cc, fj,fmaxj, gsize*sqrt(2), false);

    if ddn>1e-4 then
    begin
        if ddn>0.5 then ddn:=0.5;
        ddn0:=SImg_GetQuantile(cc,fj,ddn/3);
        ddn1:=SImg_GetQuantile(cc,fj,ddn);
    end
    else begin
        ddn0:=-simg_large;
        ddn1:=-simg_large;
    end;

    for i:=0 to fmaxi.size-1 do
        fmaxi.pt[i]-=fmini.pt[i];
    if dn>1e-3 then dni:=SImg_GetQuantile(cc,fmaxi,0.5)
    else dni:=0;

    for i:=0 to fmaxj.size-1 do
        fmaxj.pt[i]-=fminj.pt[i];
    if dn>1e-3 then dnj:=SImg_GetQuantile(cc,fmaxj,0.5)
    else dnj:=0;

    tn:=cc.threads;
    for i:=0 to tn-1 do
        dothread SImg_DeconvDenoise_constraints_thread(i,tn, fi,fmini,fmaxi,fj,fminj,fmaxj, psfw, dn,dni,dnj, ddn0,ddn1,bdn);
    threadwait;

//ifdef[debug_deconv]
//    SImg_WriteImageFile(cc,'deconvdenoise_min.fits',fmin);
//    SImg_WriteImageFile(cc,'deconvdenoise_max.fits',fmax);
//endif
    SImg_Free(cc,fmaxj);
    SImg_Free(cc,psfw);

    { extra bg constraint }
    if bg>0.099999  then
    begin
        SImg_FBGFilter(cc, fminj,fminj, sqrt(2*sqr(ifwhm)-sqr(ofwhm))*FWHM_TO_RAD, true, true);
        for i:=0 to fminj.size-1 do
        begin
            r:=fminj.pt[i];
            if fmini.pt[i]<r then fmini.pt[i]:=r;
            if fj.pt[i]<r then fj.pt[i]:=r;
        end;
    end;
    SImg_Free(cc,fminj);

    { deconvolution }
    if dc then SImg_DeconvWiener(cc, fi,fj,fmini,fmaxi, FWHM_TO_RAD*sqrt(sqr(ifwhm)-sqr(ofwhm))/bg, snr, cit);

    SImg_Free(cc,fmini);
    SImg_Free(cc,fmaxi);
    SImg_Move(cc, fj,fo);

    SImg_DecDebugLevel;
end;


