{  simg -- image processing library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{*************************************************************************}
{******* image type independent routines *********************************}
{*************************************************************************}
{*************************************************************************}
const SImg_ImageTableSize : longint = 1024;

type TSImg_ImageTableEntry=record
      pt   : pointer;
      size : int64;
      end;

var SImg_ImageTable : ^TSImg_ImageTableEntry;

{*************************************************************************}
{******* SImg_GetImageTableEntry *****************************************}
{*************************************************************************}
function SImg_GetImageTableEntry(pt:pointer;dsize:int64):longint;
var i : longint;
begin
    SImg_Sem_Lock(SImg_Sem_Img);
    try
        if (dsize<=0) or (pt=nil) then i:=SImg_ImageTableSize
        else i:=0;
        while (i<SImg_ImageTableSize) and ((SImg_ImageTable[i].pt<>pt) or (SImg_ImageTable[i].size<>dsize)) do
            i+=1;
        if i=SImg_ImageTableSize then result:=-1
        else result:=i;
    finally
        SImg_Sem_Unlock(SImg_Sem_Img);
    end;
end;

{*************************************************************************}
{******* SImg_GetEmptyImageTableEntry ************************************}
{*************************************************************************}
function SImg_GetEmptyImageTableEntry:longint;
var i,j : longint;
    tmp : ^TSImg_ImageTableEntry;
begin
    SImg_Sem_Lock(SImg_Sem_Img);
    try
        i:=0;
        while (i<SImg_ImageTableSize) and ((SImg_ImageTable[i].pt<>nil) or (SImg_ImageTable[i].size<>0)) do
            i+=1;
        if i>=SImg_ImageTableSize then
        begin
            tmp:=SImg_ImageTable;
            getmem2(SImg_ImageTable,sizeof(TSImg_ImageTableEntry)*SImg_ImageTableSize*2);
            move(tmp^,SImg_ImageTable^,sizeof(TSImg_ImageTableEntry)*SImg_ImageTableSize);
            freemem2(tmp,sizeof(TSImg_ImageTableEntry)*SImg_ImageTableSize);
            SImg_ImageTableSize*=2;
            for j:=i to SImg_ImageTableSize-1 do
            begin
                SImg_ImageTable[i].pt:=nil;
                SImg_ImageTable[i].size:=0;
            end;
        end;
        SImg_GetEmptyImageTableEntry:=i;
    finally
        SImg_Sem_Unlock(SImg_Sem_Img);
    end;
end;


{*************************************************************************}
{******* SImg_InitImgSys *************************************************}
{*************************************************************************}
procedure SImg_InitImgSys;
var i: longint;
begin
    getmem2(SImg_ImageTable,SImg_ImageTableSize*sizeof(TSImg_ImageTableEntry));
    for i:=0 to SImg_ImageTableSize-1 do
    begin
        SImg_ImageTable[i].pt:=nil;
        SImg_ImageTable[i].size:=0;
    end;
end;


{*************************************************************************}
{******* SImg_DoneImgSys *************************************************}
{*************************************************************************}
procedure SImg_DoneImgSys(ec:longint);
var i,j : longint;
    cc  : CSImg_Config;
begin
    cc:=SImg_Config_Default;
    j:=0;
    for i:=0 to SImg_ImageTableSize-1 do
        with SImg_ImageTable[i] do
        begin
            if (pt<>nil) and (size>0) then
            begin
                j+=1;
//ifdef[debug]
                if (ec=0) then cc.debug('SImg_DoneImgSys: Free '+base1024str(size)+' Bytes for Img(id='+int2str(i)+')'+' at '+ptr2str(pt));
//endif
                freemem2(pt,size);
            end;
        end;
    if (ec=0) and (j>0) then cc.warning('SImg_DoneImgSys: deallocating memory for '+int2str(j)+' Images');
    freemem2(SImg_ImageTable,SImg_ImageTableSize*sizeof(TSImg_ImageTableEntry));
end;

