{  simg -- image processing library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

// define[debug_interpolate]

{*************************************************************************}
{******* SImg_InterpolateMasked ******************************************}
{*************************************************************************}
{******* SImg_InterpolateMasked_thread ***********************************}
{ Interpolate masked pixels
  fi      : input and output
  mask    : >0 : interpolate
            128: interpolate + border
  pm      : mask pattern
}
threadprocedure SImg_InterpolateMasked_thread(fi:TSImg_FPImage; rm:TSImg_RDImage; ya:longint; yz:longint);
const rsqrt2 = sqrt(0.5);
      rsqrt5 = sqrt(0.2);
var hist,hist_w,hist_v : array[0..7] of simgfp;
    x,y,i,j,k          : longint;
    r,s,t              : double;
    d                  : ^simgfp;
    b,c                : ^byte;
//define[search(][,#1,#2);][
    begin
        j:=(#1)*rm.nc;
        k:=1;
        c:=b+j;
        while c^ and 128=0 do
        begin
            c+=j;
            k+=1;
        end;
        if c^ and 127=0 then
        begin
            hist[#0]:=d[k*j];
            r:=#2/k;
        end else
        begin
            hist[#0]:=d^;
            r:=1e-12;
        end;
    end else
    begin
        hist[#0]:=d^;
        r:=1e-12;
    end;
    hist_w[#0]:=r;
    s+=r;
    if r>t then t:=r;
]

begin
    d:=fi.pt + ya*fi.sl;
    b:=rm.ptb + ya*rm.sl;
    hist[0]:=0;     // makes the compiler happy
    hist_v[0]:=0;   // makes the compiler happy
    for y:=ya to yz do
        for x:=0 to fi.nx-1 do
            for i:=0 to fi.nc-1 do
            begin
                if b^ and 127<>0 then
                begin
                    s:=0; t:=0;
                    if (y>0) and (x>0)             then search(0, -rm.nx-1, rsqrt2);
                    if (y>0)                       then search(1, -rm.nx  , 1);
                    if (y>0) and (x<fi.nx-1)       then search(2, -rm.nx+1, rsqrt2);
                    if (x<fi.nx-1)                 then search(3, 1       , 1);
                    if (y<fi.ny-1) and (x<fi.nx-1) then search(4, rm.nx+1 , rsqrt2);
                    if (y<fi.ny-1)                 then search(5, rm.nx   , 1);
                    if (y<fi.ny-1) and (x>0)       then search(6, rm.nx-1 , rsqrt2);
                    if (x>0)                       then search(7, -1      , 1);
                    s*=0.5;

                    { sort }
                    for j:=0 to high(hist) do
                    begin
                        r:=hist[j];
                        k:=j;
                        while (k>0) and (hist[k-1]>r) do
                        begin
                            hist[k]:=hist[k-1];
                            hist_v[k]:=hist_v[k-1];
                            k-=1;
                        end;
                        hist[k]:=r;
                        hist_v[k]:=hist_w[j];
                    end;

                    { median }
                    j:=1;
                    r:=hist_v[0]+hist_v[1];
                    while (j<high(7)) and (r<s) do
                    begin
                        j+=1;
                        r+=hist_v[j];
                    end;
                    r:=(r-s)/hist_v[j];
                    if r>1 then r:=1;
                    d^:=hist[j-1]*r + hist[j]*(1-r);
                    if t<0.0079 then b^:=b^ or 127
                    else b^:=(b^ and 128) or round(1+1/t);
                end;
                d+=1;
                b+=1;
            end;
end;

//udefine[search(]

{******* SImg_InterpolateMasked ******************************************}
{ Interpolate masked pixels
  fi      : input and output
  mask    : interpolate if mask and pm<>0
  pm      : mask pattern
  fd      : filter radius divider relative to nearest interpolation point, as larger as less low pass filtering. Disabled if >127.
}
procedure SImg_InterpolateMasked(cc:CSImg_Config; var fi:TSImg_FPImage; mask:TSImg_RDImage; pm,fd:byte);
var x,y,c,i     : longint;
    rm          : TSImg_RDImage;
    by          : boolean;
    r           : double;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_InterpolateMasked: fi='+SImg_InfoStr2(fi)+' mask='+SImg_InfoStr2(mask)+' pm='+int2str(pm)+' fd='+int2str(fd));
    if not SImg_Recover(cc,fi) then cc.error('SImg_InterpolateMasked: Invalid image: fi');
    if not SImg_Recover(cc,mask) then cc.error('SImg_InterpolateMasked: Invalid image: mask');
    if mask.fmt<>SImg_RawByte then cc.error('SImg_InterpolateMasked: Invalid data format of mask');
    if mask.size<>fi.size then cc.error('SImg_InterpolateMasked: Invalid size of mask');
    if fd<1 then fd:=1;

    { internal mask
        128     : pixel defined
        >128    : undefined boundary pixel
        1..127  : undefined internal pixel,
        bits 0 to 6 are used to store distance to closest defined pixel
    }
    rm.pt:=nil;
    SImg_New(cc, rm, SImg_RawByte, mask.nx,mask.ny,mask.nc, false);
    i:=0;
    for y:=0 to mask.ny-1 do
    begin
        by:=(y=0) or (y=mask.ny-1);
        for x:=0 to mask.nx-1 do
            for c:=0 to mask.nc-1 do
            begin
                if mask.ptb[i] and pm=0 then rm.ptb[i]:=128
                else if by or (x=0) or (x=mask.nx-1) then rm.ptb[i]:=129
                else rm.ptb[i]:=1;
                i+=1;
            end;
    end;

    { interpolation }
    c:=cc.threads;
    x:=0;
    for i:=0 to c-1 do
    begin
        y:=fi.ny*(i+1) div c;
        dothread SImg_InterpolateMasked_thread(fi,rm, x,y-1);
        x:=y;
    end;
    threadwait;

    { lp filtering }
    c:=0;
    if fd<128 then
        for i:=0 to rm.size-1 do
        begin
            x:=rm.ptb[i] and 127;
            if x>c then c:=x;
        end;

    if c>fd then
    begin
        r:=255/(c-fd);
        for i:=0 to rm.size-1 do
        begin
            x:=rm.ptb[i] and 127;
            if x>fd then rm.ptb[i]:=255-round((x-fd)*r)
            else rm.ptb[i]:=255;
        end;
//ifdef[debug_interpolate]
        cc.info('SImg_InterpolateMasked: max interpolation distance is '+int2str(c));
        SImg_WriteImageFile(cc,'interpolateMasked-rm.fits',rm);
//endif[debug_interpolate]
        SImg_Filter(cc, fi,fi, c/fd-1, rm);
    end;
    SImg_Free(cc, rm);

    SImg_DecDebugLevel;
end;

procedure SImg_InterpolateMasked(cc:CSImg_Config; var fi:TSImg_FPImage; mask:TSImg_RDImage; pm:byte);
begin
    SImg_InterpolateMasked(cc, fi, mask, pm, 3);
end;


{*************************************************************************}
{******* SImg_InterpolateMasked2 *****************************************}
{*************************************************************************}
{ Interpolate masked pixels. SImg_InterpolateMasked is usually better due to median filter.
  fi      : input and output
  mask    : mask and pm <> 0 : interpolate
  pm      : mask pattern
}
procedure SImg_InterpolateMasked2(cc:CSImg_Config; fi:TSImg_FPImage; mask:TSImg_RDImage; pm:byte);
const  sfact  = 2;      { scaling factor }
       wfact  = 10;     { weight factor }
var fm,fn  : TSImg_FPImage;
    i,j    : longint;
    s      : double;
    a,b    : simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_InterpolateMasked: fi='+SImg_InfoStr2(fi)+' mask='+SImg_InfoStr2(mask)+' pm='+int2str(pm));
    if not SImg_Recover(cc,fi) then cc.error('SImg_InterpolateMasked: Invalid image: fi');
    if not SImg_Recover(cc,mask) then cc.error('SImg_InterpolateMasked: Invalid image: mask');
    if mask.fmt<>SImg_RawByte then cc.error('SImg_InterpolateMasked: Invalid data format of mask');
    if mask.size<>fi.size then cc.error('SImg_InterpolateMasked: Invalid size of mask');

    { temporary images }
    fm.pt:=nil;
    SImg_New(cc, fm, fi.nx,fi.ny,fi.nc*2, false);
    j:=0;
    for i:=0 to fi.size-1 do
        if mask.ptb[i] and pm=0 then
        begin
            fm.pt[i*2  ]:=fi.pt[i]*1;
            fm.pt[i*2+1]:=1;
        end
        else begin
            fm.pt[i*2  ]:=0;
            fm.pt[i*2+1]:=0;
            j+=1;
        end;

//ifdef[debug_interpolate]
    cc.info('SImg_InterpolateMasked: '+int2str(j)+' undefined pixels');
//endif
    s:=0.5;
    fn.pt:=nil;
    while (j>0) and (s<fi.nx) and (s<fi.ny) do
    begin
        s*=sfact;
        SImg_Filter(cc, fm,fn,s);
        j:=0;
        for i:=0 to fi.size-1 do
        begin
            a:=fm.pt[i*2+1];
            if a<1 then
            begin
                b:=fn.pt[i*2+1]*wfact;
                if a+b>1 then
                begin
                    fm.pt[i*2]+=fn.pt[i*2]*wfact*(1-a)/b;
                    fm.pt[i*2+1]:=1;    { avoids rounding errors }
                end
                else begin
                    fm.pt[i*2]+=fn.pt[i*2]*wfact;
                    fm.pt[i*2+1]:=a+b;
                    j+=1;
                end;
            end;
        end;
//ifdef[debug_interpolate]
//      SImg_WriteImageFile(cc,'interpolateMasked2-'+int2str(round(s))+'.fits',fm);
        cc.info('SImg_InterpolateMasked: size='+real2str(s,-1)+': '+int2str(j)+' undefined pixels');
//endif
    end;
    SImg_Free(cc,fn);

    for i:=0 to fi.size-1 do
        if mask.ptb[i] and pm<>0 then fi.pt[i]:=fm.pt[i*2];
    SImg_Free(cc,fm);

    SImg_DecDebugLevel;
end;

