{  simg -- visualization library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

const siv_visual_list : CSIV_Visual = nil;
      siv_dont_stop : boolean = false;

{ ************************************************************************* }
{ ******** visual events ************************************************** }
{ ************************************************************************* }
{ ******** siv_visual_destroy ********************************************* }
procedure siv_visual_destroy(widget : PGtkWidget; visual: CSIV_Visual); cdecl;
var tmp    : ^CSIV_Visual;
    fi     : TSImg_FPImage;
begin
    siv_debug_begin('siv_visual_destroy');
    tmp:=@siv_visual_list;
    while (tmp^<>nil) and (tmp^<>visual) do
        tmp:=@tmp^.next;
    if tmp^=nil then SIV_Error('siv_visual_destroy: Trying to destroy unknown window')
    else begin
        if tmp^.window<>widget then siv_warning('siv_visual_destroy: Widget do not match with visual.window');
        tmp^:=visual.next;
        if (siv_visual_list=nil) and SImg_Recover(siv_config,visual.image.img16) then
        begin
            fi.pt:=nil;
            CSIV_Visual.init(siv_load_nothing, fi, '',
            visual.dir, visual.history,
            gtk_widget_visible(visual.controlbar), gtk_widget_visible(visual.histbar), visual.combinemte,
                nil);
        end;
        visual.done;
    end;
    if (siv_visual_list=nil) and (not siv_dont_stop) then
    begin
         siv_message_quit;
        gtk_main_quit();
        end;
    siv_dont_stop:=false;
    siv_debug_end;
end;

{ ******** siv_visual_reload ********************************************** }
procedure siv_visual_reload(widget:PGtkWidget; self:CSIV_Visual); cdecl;
var fi  : TSIMG_FPImage;
begin
    if self=nil then
    begin
        siv_warning('siv_visual_clone: self=nil');
        exit;
    end;
    with self do
    begin
        siv_debug_begin('siv_visual_clone: anc='+image.title);
        try
            fi.pt:=nil;
            if (image.name<>'') and (FileExists(image.name)) then
            begin
                SImg_ReadImageFile(siv_config,image.name,fi);
                if SImg_Recover(siv_config, fi) then
                begin
                    if (fi.nx<>image.img16.nx) or (fi.ny<>image.img16.ny) or(fi.nc<>image.img16.nc) then
                    begin
                        siv_warning('siv_visual_reload: Unexpected image dimensions');
                        CSIV_Visual.init(siv_load_fi, fi, image.name,
                            '', history,
                            gtk_widget_visible(controlbar), gtk_widget_visible(histbar), combinemte,
                            nil);
                    end
                    else CSIV_Visual.init(siv_load_fi, fi, image.name,
                            '', history,
                            gtk_widget_visible(controlbar), gtk_widget_visible(histbar), combinemte,
                            self);
                 end;
                SImg_Free(siv_config, fi);
            end
            else if SImg_Recover(siv_config,image.img16) then
            begin
                siv_warning('siv_visual_reload: Can''t reload from stdin or file does not exist, clone function is used instead');
                CSIV_Visual.init(siv_load_clone, fi, image.name,
                    '', history,
                    gtk_widget_visible(controlbar), gtk_widget_visible(histbar), combinemte,
                    self);
            end;
        except
            SIV_OutputCatchedException;
        end;
        siv_debug_end;
    end;
end;

{ ******** siv_visual_clone *********************************************** }
procedure siv_visual_clone(widget:PGtkWidget; self:CSIV_Visual); cdecl;
var fi  : TSIMG_FPImage;
begin
    if self=nil then
    begin
        siv_warning('siv_visual_clone: self=nil');
        exit;
    end;
    with self do
    begin
        siv_debug_begin('siv_visual_clone: anc='+image.title);
        try
            if SImg_Recover(siv_config,image.img16) then
            begin
                fi.pt:=nil;
                CSIV_Visual.init(siv_load_clone, fi, image.name,
                    '', history,
                    gtk_widget_visible(controlbar), gtk_widget_visible(histbar), combinemte,
                    self);
            end;
        except
            SIV_OutputCatchedException;
        end;
        siv_debug_end;
    end;
end;

{ ******** siv_visual_fit ************************************************* }
procedure siv_visual_fit(widget:PGtkWidget; self:CSIV_Visual); cdecl;
begin
    if self=nil then
    begin
        siv_warning('siv_visual_fit: self=nil');
        exit;
    end;
    self.fit;
end;

{ ******** siv_visual_combinemte ****************************************** }
procedure siv_visual_combinemte(widget:PGtkWidget; self:CSIV_Visual); cdecl;
begin
    if self=nil then
    begin
        siv_warning('siv_visual_combinemte: self=nil');
        exit;
    end;
    with self do
    begin
        combinemte:=true;
        gtk_widget_hide(bcombinemte);
        gtk_widget_show(bnocombinemte);
    end;
end;

{ ******** siv_visual_combinemte ****************************************** }
procedure siv_visual_nocombinemte(widget:PGtkWidget; self:CSIV_Visual); cdecl;
begin
    if self=nil then
    begin
        siv_warning('siv_visual_nocombinemte: self=nil');
        exit;
    end;
    with self do
    begin
        combinemte:=false;
        gtk_widget_hide(bnocombinemte);
        gtk_widget_show(bcombinemte);
    end;
end;

{ ******** info window *************************************************** }
{ ******** siv_visual_info_destroy *************************************** }
procedure siv_visual_info_destroy(widget:PGtkWidget; self:CSIV_Visual); cdecl;
begin
    if self=nil then
    begin
        siv_warning('siv_visual_info_destroy: self=nil');
        exit;
    end;
    self.info_window:=nil;
end;

{ ******** siv_visual_info *********************************************** }
procedure siv_visual_info_add(vbox:PGtkWidget; first:boolean; header,left,right:simgstring);
var text_widget,hbox   : PGtkWidget;
    s                  : simgstring;
begin
    setlength(s,length(header));
    fillchar(s[1],length(header),'_');
    if not first then
    begin
        header:=#10+header; s:=' '+s;
    end;
    text_widget:=gtk_label_new(str2pchar(header));
    gtk_label_set_pattern(PGtkLabel(text_widget),str2pchar(s));
    gtk_box_pack_start(GTK_BOX(vbox),text_widget,false,false,0);
    gtk_widget_show(text_widget);

    hbox:=gtk_hbox_new(false,4);
    gtk_box_pack_start(GTK_BOX(vbox),hbox,false,false,0);
    gtk_widget_show(hbox);

    text_widget:=gtk_label_new(str2pchar(left));
    gtk_label_set_justify(GTK_LABEL(text_widget),GTK_JUSTIFY_LEFT);
    gtk_box_pack_start(GTK_BOX(hbox),text_widget,false,false,0);
    gtk_widget_show(text_widget);

    text_widget:=gtk_label_new(str2pchar(right));
    gtk_label_set_justify(GTK_LABEL(text_widget),GTK_JUSTIFY_LEFT);
    gtk_box_pack_start(GTK_BOX(hbox),text_widget,false,false,0);
    gtk_widget_show(text_widget);
end;

procedure siv_visual_info(widget:PGtkWidget; self:CSIV_Visual); cdecl;
var scroll,align,ivbox  : PGtkWidget;
    ivbox0,close_b      : PGtkWidget;
    req1,req2           : TGtkRequisition;
    title               : simgstring;
begin
    if self=nil then
    begin
        siv_warning('siv_visual_info: self=nil');
        exit;
    end;
    with self do
        if ((info_window=nil) or (not gtk_is_window(info_window))) and SImg_Recover(siv_config,image.img16) then
        begin
            title:='Info: '+image.title;
            info_window:=gtk_window_new(GTK_WINDOW_TOPLEVEL);
            gtk_window_set_title(PGtkWindow(info_window), str2pchar(title));
            g_signal_connect(GTK_OBJECT(info_window), 'destroy', GTK_SIGNAL_FUNC(@siv_visual_info_destroy), self);
            gtk_widget_show(info_window);

            ivbox0:=gtk_vbox_new(false,0);
            gtk_container_add(PGtkContainer(info_window), ivbox0);
            gtk_widget_show(ivbox0);

            scroll:=gtk_scrolled_window_new(nil,nil);
            gtk_scrolled_window_set_policy(PGtkScrolledWindow(scroll), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
            gtk_box_pack_start(GTK_BOX(ivbox0),scroll,true,true,0);
            gtk_widget_show(scroll);

            align:=gtk_alignment_new(0.5,0.5,0,0);
            gtk_scrolled_window_add_with_viewport(PGtkScrolledWindow(scroll),align);
            gtk_widget_show(align);

            ivbox:=gtk_vbox_new(false,0);
            gtk_container_add(PGtkContainer(align), ivbox);
            gtk_widget_show(ivbox);

            siv_visual_info_add(ivbox, true, 'General:',
                'Dimensions (X x Y x C)'+#10+
                'Minimum'+#10+
                'Maximum'+#10+
                'Mean value'+#10+
                'Mean deviation'+#10+
                'Standard deviation (RMS)',
                ': '+int2str(image.img16.nx)+' x '+int2str(image.img16.ny)+' x '+int2str(image.img16.nc)+#10+
                ': '+real2str(image.minv,6)+#10+
                ': '+real2str(image.maxv,6)+#10+
                ': '+real2str(image.mean,6)+#10+
                ': '+real2str(image.meandev,5)+#10+
                ': '+real2str(image.rms,5));

            { close }
            close_b:=siv_button_new_with_label('Close');
            gtk_box_pack_start(GTK_BOX(ivbox0),close_b,false,false,0);
            g_signal_connect_swapped(GTK_OBJECT(close_b),'clicked',GTK_SIGNAL_FUNC(@gtk_widget_destroy),GTK_OBJECT(info_window));
            gtk_widget_show(close_b);
            siv_tooltip_new(close_b,'Close info dialog',nil);

            gtk_widget_size_request(ivbox, @req1);
            gtk_widget_size_request(close_b, @req2);
            gdk_window_resize(info_window^.window, req1.width+30, (req1.height+req2.height)+30);
        end;
end;


{ ******** open file ****************************************************** }
{ ******** siv_visual_openfile ******************************************** }
procedure siv_visual_openfile(widget:PGtkWidget; self:CSIV_Visual); cdecl;
var s      : simgstring;
    dialog : PGtkWidget;
begin
    if self=nil then
    begin
        siv_warning('siv_visual_openfile: self=nil');
        exit;
    end;
    s:='Open image in new window';
    with self do
    begin
        dialog:=gtk_file_chooser_dialog_new(str2pchar(s), PGtkWindow(window), GTK_FILE_CHOOSER_ACTION_OPEN, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);
        gtk_file_chooser_set_current_folder(PGtkFileChooser(dialog),str2pchar(dir));
        if gtk_dialog_run(PGtkDialog(dialog))=GTK_RESPONSE_ACCEPT then
        begin
            s:=gtk_file_chooser_get_filename(PGtkFileChooser(dialog));
            opennew(s);
        end;
    end;
    gtk_widget_destroy(dialog);
end;

{ ******** history ******************************************************** }
{ ******** siv_visual_history_activate ************************************ }
procedure siv_visual_history_activate(widget : PGtkWidget; self:CSIV_Visual); cdecl;
var pc     : ^char;
begin
    if self=nil then
    begin
        siv_warning('siv_visual_history_activate: self=nil');
        exit;
    end;
    with self do
    begin
        pc:=gtk_combo_box_get_active_text(PGtkComboBox(histcombo));
        opennew(pc);
    end;
end;

{ ******** histbar ******************************************************** }
{ ******** siv_visual_show_histbar **************************************** }
procedure siv_visual_show_histbar(widget:PGtkWidget; self:CSIV_Visual); cdecl;
var requisition  : TGtkRequisition;
    x,y          : longint;
begin
    if self=nil then
    begin
        siv_warning('siv_visual_show_histbar: self=nil');
        exit;
    end;
    with self do
    begin
        image.history_markers(marker_x,marker_y);
        siv_draw_redraw(PSivDraw(histl));
        siv_draw_redraw(PSivDraw(histr));
        gtk_widget_show(histbar);
        gtk_widget_hide(bshow_hb);
        gtk_widget_show(bhide_hb);
        gtk_widget_size_request(histbar,@requisition);
        gtk_window_get_size(PGtkWindow(window), @x, @y);
        gtk_window_resize(PGtkWindow(window), x, y+requisition.height);
    end;
end;

{ ******** siv_visual_hide_histbar **************************************** }
procedure siv_visual_hide_histbar(widget:PGtkWidget; self:CSIV_Visual); cdecl;
var requisition  : TGtkRequisition;
    x,y          : longint;
begin
    if self=nil then
    begin
        siv_warning('siv_visual_hide_histbar: self=nil');
        exit;
    end;
    with self do
    begin
        gtk_widget_hide(histbar);
        gtk_widget_hide(bhide_hb);
        gtk_widget_show(bshow_hb);
        gtk_widget_size_request(histbar,@requisition);
        gtk_window_get_size(PGtkWindow(window), @x, @y);
        gtk_window_resize(PGtkWindow(window), x, y-requisition.height);
    end;
end;


{ ******** controlbar ***************************************************** }
{ ******** siv_visual_show_controlbar ************************************* }
procedure siv_visual_show_controlbar(widget:PGtkWidget; self:CSIV_Visual); cdecl;
var requisition  : TGtkRequisition;
    x,y          : longint;
begin
    if self=nil then
    begin
        siv_warning('siv_visual_show_controlbar: self=nil');
        exit;
    end;
    with self do
    begin
        image.history_markers(marker_x,marker_y);
        siv_draw_redraw(PSivDraw(histl));
        siv_draw_redraw(PSivDraw(histr));
        gtk_widget_show(controlbar);
        gtk_widget_hide(bshow_cb);
        gtk_widget_show(bhide_cb);
        gtk_widget_size_request(controlbar,@requisition);
        gtk_window_get_size(PGtkWindow(window), @x, @y);
        gtk_window_resize(PGtkWindow(window), x, y+requisition.height);
    end;
end;

{ ******** siv_visual_hide_controlbar ************************************* }
procedure siv_visual_hide_controlbar(widget:PGtkWidget; self:CSIV_Visual); cdecl;
var requisition  : TGtkRequisition;
    x,y          : longint;
begin
    if self=nil then
    begin
        siv_warning('siv_visual_hide_controlbar: self=nil');
        exit;
    end;
    with self do
    begin
        gtk_widget_hide(controlbar);
        gtk_widget_hide(bhide_cb);
        gtk_widget_show(bshow_cb);
        gtk_widget_size_request(controlbar,@requisition);
        gtk_window_get_size(PGtkWindow(window), @x, @y);
        gtk_window_resize(PGtkWindow(window), x, y-requisition.height);
    end;
end;

{ ******** marker ********************************************************* }
{ ******** siv_pos2str **************************************************** }
function siv_pos2str(x,y:longint):simgstring;
var s  : shortstring;
begin
    str(x,s);
    result:='X='+s;
    str(y,s);
    result+='  Y='+s;
end;

{ ******** siv_poscolor2str **************************************************** }
function siv_poscolor2str(image:CSIV_Image; x,y:longint):simgstring;
var i    : longint;
    t,r  : double;
    d    : ^word;
begin
    if (x>=0) and (y>=0) and (x<image.img16.nx) and (y<image.img16.ny) and (image.img16.pt<>nil) then
    begin
        r:=1/65535;
        d:=image.img16.ptw + y*image.img16.sl + x*image.img16.nc;
        result:='C=(';
        for i:=0 to image.img16.nc-1 do
        begin
            t:=d^*r/(1-image.delta+image.delta*d^*r)*(image.maxv-image.minv) + image.minv;
            if i=0 then result+=real2str(t,5)
            else result+=', '+real2str(t,5);
            d+=1;
        end;
        result+=')';
    end
    else result:='';
end;

{ ******** siv_visual_marker_button_press ********************************* }
function siv_visual_marker_button_press(widget:PGtkWidget;event:PGdkEventButton; self:CSiv_Visual):longint; cdecl;
var draw  : PSIVDraw;
    x,y   : longint;
begin
    result:=1;
    if event=nil then
    begin
        siv_warning('siv_visual_marker_button_press: event=nil');
        exit;
    end;
    if self=nil then
        begin
        siv_warning('siv_visual_marker_button_press: self=nil');
        exit;
    end;
    with self do
    begin
        draw:=PSIVDraw(iw.draw);
        if event^.button=1 then
        begin
            x:=round(event^.x)+draw^.last_x;
            y:=round(event^.y)+draw^.last_y;
            if image.scale<0 then setmarker(x div (1-image.scale), y div (1-image.scale))
            else if image.scale=0 then setmarker(x, y)
            else setmarker(x*(1+image.scale)+(1+image.scale) shr 1, y*(1+image.scale)+(1+image.scale) shr 1);
        end;
    end;
end;

{ ******** siv_visual_marker_motion *************************************** }
function siv_visual_marker_motion(widget:PGtkWidget; event:PGdkEventMotion; self:CSIV_Visual):longint; cdecl;
var draw      : PSIVDraw;
    s         : simgstring;
    x,y       : longint;
    state     : TGdkModifierType;
begin
    result:=1;
    if event=nil then
    begin
        siv_warning('siv_visual_marker_motion_event: event=nil');
        exit;
    end;
    if self=nil then
    begin
        siv_warning('siv_iw_motion_event: self=nil');
        exit;
    end;
    with self do
    begin
        draw:=PSIVDraw(iw.draw);
        if draw^.rgb=nil then exit;
        if event^.is_hint>0 then gdk_window_get_pointer(event^.window, @x, @y, @state)
        else begin
            x:=round(event^.x);
            y:=round(event^.y);
            state:=event^.state;
        end;

        x+=draw^.last_x;
        y+=draw^.last_y;
        if image.scale<0 then
        begin
            x:=x div (1-image.scale);
            y:=y div (1-image.scale);
        end
        else if image.scale>0 then
        begin
            x:=x*(1+image.scale)+(1+image.scale) shr 1;
            y:=y*(1+image.scale)+(1+image.scale) shr 1;
        end;
        s:='Mouse: ';
        if (x>=0) and (y>=0) and (x<image.img16.nx) and (y<image.img16.ny)
        then s+=siv_pos2str(x,y);
        gtk_label_set_text(PGtkLabel(pointer_label),str2pchar(s));

        if state and GDK_BUTTON1_MASK>0 then setmarker(x,y);
    end;
end;

{ ******** siv_visual_marker_key_press ************************************ }
function siv_visual_marker_key_press(widget:PGtkWidget; event:PGdkEventKey; self:CSiv_Visual):longint; cdecl;
var x,y   : longint;
begin
    result:=1;
    if event=nil then
    begin
        siv_warning('siv_visual_marker_key_press: event=nil');
        exit;
    end;
    if self=nil then
    begin
        siv_warning('siv_visual_marker_key_press: self=nil');
        exit;
    end;
    with self do
    begin
        x:=marker_x;
        y:=marker_y;
        case event^.keyval of
            GDK_key_Left  : x-=1;
            GDK_key_Right : x+=1;
            GDK_key_Up    : y-=1;
            GDK_key_Down  : y+=1;
        end;
        if (x<>marker_x) or (y<>marker_y) then setmarker(x,y);
    end;
end;

{ ******** siv_visual_marker_focus_in ************************************* }
function siv_visual_marker_focus_in(widget:PGtkWidget; event:PGdkEventFocus; self:CSiv_Visual):longint; cdecl;
begin
    result:=1;
    if widget=nil then
    begin
        siv_warning('siv_visual_marker_focus_in: widget=nil');
        exit;
    end;
    gtk_widget_set_flags(widget,GTK_HAS_FOCUS);
end;

{ ******** siv_visual_marker_focus_out ************************************ }
function siv_visual_marker_focus_out(widget:PGtkWidget; event:PGdkEventFocus; self:CSiv_Visual):longint; cdecl;
begin
    result:=1;
    if widget=nil then
    begin
        siv_warning('siv_visual_marker_focus_out: widget=nil');
        exit;
    end;
    gtk_widget_unset_flags(widget,GTK_HAS_FOCUS);
end;


{ ******** scale ********************************************************** }
{ ******** siv_scale2str ************************************************** }
function siv_scale2str(scale:longint):simgstring;
begin
    str(abs(scale)+1,result);
    if scale<0 then result:=result+':1'
    else result:='1:'+result;
end;

{ ******** siv_scale_update *********************************************** }
procedure siv_scale_update(widget:PGtkWidget; self:CSIV_Visual); cdecl;
var i   : longint;
begin
    if self=nil then
    begin
        siv_warning('siv_scale_update: self=nil');
        exit;
    end;
    with self do
    begin
        i:=round(scale_scale.adj^.value);
        if image.scale<>i then
        begin
            image.scale:=-i;
            image.update_scale:=true;
            redraw;
            scale_scale.set_label(-1,'Scale='+siv_scale2str(image.scale));
            if (marker_x>=0) and (marker_y>=0) then
            begin
                i:=marker_x;
                marker_x:=-1;
                setmarker(i,marker_y);
            end;
        end;
    end;
end;

{ ******** other slider *************************************************** }
{ ******** siv_limits_update ********************************************** }
procedure siv_limits_update(widget:PGtkWidget; self:CSIV_Visual); cdecl;
begin
    if self=nil then
    begin
        siv_warning('siv_limits_update: self=nil');
        exit;
    end;
    with self do
    begin
        image.ll:=ll_scale.adj^.value;
        image.ul:=ul_scale.adj^.value;
        image.update_limits:=true;
        redraw;
        ll_scale.set_label(-1,'ql='+siv_real2str(image.ql*100,3)+'%');
        ul_scale.set_label(-1,'qu='+siv_real2str(image.qu*100,3)+'%');
        ll_scale.stepcheck;
        ul_scale.stepcheck;
    end;
end;


{ ******** siv_bg_update ************************************************** }
procedure siv_bg_update(widget:PGtkWidget; self:CSIV_Visual); cdecl;
begin
    if self=nil then
    begin
        siv_warning('siv_bg_update: self=nil');
        exit;
    end;
    with self do
    begin
        image.bet:=bet_scale.adj^.value;
        image.gam:=gam_scale.adj^.value;
        image.update_bg:=true;
        redraw;
        bet_scale.set_label(-1,'Beta='+siv_real2str(image.bet,2));
        gam_scale.set_label(-1,'Gamma='+siv_real2str(image.gam,2));
        bet_scale.stepcheck;
        gam_scale.stepcheck;
    end;
end;


{ ******** toggle buttons ************************************************* }
{ ******** siv_inv_update ************************************************* }
procedure siv_inv_update(widget:PGtkWidget; self:CSIV_Visual); cdecl;
var b  : boolean;
begin
    if self=nil then
    begin
        siv_warning('siv_inv_update: self=nil');
        exit;
    end;
    with self do
    begin
        b:=gtk_toggle_button_get_active(PGtkToggleButton(inv_tb));
        if image.inv<>b then
        begin
            image.inv:=b;
            image.update_bg:=true;
            redraw;
        end;
    end;
end;

{ ******** siv_swap_update ************************************************* }
procedure siv_swap_update(widget:PGtkWidget; self:CSIV_Visual); cdecl;
begin
    if self=nil then
    begin
        siv_warning('siv_inv_update: self=nil');
        exit;
    end;
    with self do
    begin
        ll_scale.adj^.value:=image.ul;
        ul_scale.adj^.value:=image.ll;
        siv_limits_update(widget,self);
        gtk_adjustment_value_changed(ll_scale.adj);
        gtk_adjustment_value_changed(ul_scale.adj);
    end;
end;

{ ******** siv_log_update ************************************************* }
procedure siv_log_update(widget:PGtkWidget; self:CSIV_Visual); cdecl;
var b  : boolean;
begin
    if self=nil then
    begin
        siv_warning('siv_log_update: self=nil');
        exit;
    end;
    with self do
    begin
        b:=gtk_toggle_button_get_active(PGtkToggleButton(log_tb));
        if image.log<>b then
        begin
            image.log:=b;
            image.redraw_histl:=true;
            image.redraw_histr:=true;
            redraw;
        end;
    end;
end;

{ ******** siv_color_update *********************************************** }
procedure siv_color_update(widget:PGtkWidget; self:CSIV_Visual); cdecl;
var i : longint;
begin
    if (widget=nil) or (not gtk_is_combo_box(widget)) then
    begin
        siv_warning('siv_color_update: widget is no gtk_combo_box');
        exit;
    end;

    if self=nil then
    begin
        siv_warning('siv_color_update: self=nil');
        exit;
    end;

    with self do
    begin
        i:=gtk_combo_box_get_active(PGtkComboBox(widget));
        if (i<0) or (i>=image.img16.nc) then
        begin
            siv_warning('siv_color_update: invalid color');
            exit;
        end;
        if widget=red_menu then image.red:=i
        else if widget=green_menu then image.green:=i
        else if widget=blue_menu then image.blue:=i
        else begin
            siv_warning('siv_color_update: unknown widget');
            exit;
        end;
        image.update_colors:=true;
        redraw;
    end;
end;


{ ************************************************************************* }
{ ******** CSIV_Visual **************************************************** }
{ ************************************************************************* }
{ ******** init *********************************************************** }
constructor CSIV_Visual.init(load:longint; fi:TSImg_FPImage; fn,adir:simgstring; h:CSIV_History; scb,shb,cmte:boolean; anc:CSIV_Visual);
var vboxl,vboxr   : PGtkWidget;  { left/right vbox in controlbar }
    hbox1         : PGtkWidget;
    vbox1         : PGtkWidget;
    eventbox      : PGtkWidget;
    tmpwid        : PGtkWidget;
    i             : longint;
    tmp           : ^CSiv_Visual;
    s             : simgstring;
    rect          : TGDKRectAngle;
begin
    if (anc<>nil) and (anc.image<>nil) then s:=anc.image.title
    else s:='none';
    siv_debug_begin('CSIV_Visual.init: fi='+SImg_InfoStr2(fi)+' fn='+fn+' anc='+s);
    dir:='';
    if adir='' then
    begin
        dir:=ExpandFileName(extractFileDir(fn));
        if dir[length(dir)]<>'/' then dir+='/';
    end
    else dir:=adir;
    if h=nil then history:=CSIV_History.init(siv_global_history)
    else history:=CSIV_History.init(h);
    history.insert(fn);

    case load of
        siv_load_fi : image:=CSIV_Image.init(fi,fn);
        siv_load_fn :
            begin
                try
                    SImg_ReadImageFile(siv_config,fn,fi);
                except
                    SIV_OutputCatchedException;
                    fi.pt:=nil;
                end;
                image:=CSIV_Image.init(fi,fn);
                SImg_Free(siv_config, fi);
            end;
    siv_load_clone :
        if (anc<>nil) and (anc.image<>nil) then image:=CSIV_Image.init(anc.image)
        else image:=CSIV_Image.init(nil);
    else
        begin
            if load<>siv_load_nothing then siv_warning('CSIV_Visual.init: Invalid load value');
            image:=CSIV_Image.init(nil);
        end;
    end;

    if anc<>nil then
    begin
        image.scale:=anc.image.scale;
        image.red:=anc.image.red;
        image.green:=anc.image.green;
        image.blue:=anc.image.blue;
        image.ll:=anc.image.ll;
        image.ul:=anc.image.ul;
        image.bet:=anc.image.bet;
        image.gam:=anc.image.gam;
        image.inv:=anc.image.inv;
        image.log:=anc.image.log;
    end;
    image.update;
    { **************** window **************** }
    window:=gtk_window_new(GTK_WINDOW_TOPLEVEL);
    if anc<>nil then
    begin
        gtk_window_get_size(PGtkWindow(anc.window), @rect.x, @rect.y);
        gtk_window_set_default_size(PGtkWindow(window), rect.x, rect.y);
    end;

    gtk_widget_show(window);
    g_signal_connect(PGTKOBJECT(window), 'destroy', GTK_SIGNAL_FUNC(@siv_visual_destroy), self);
    gtk_window_set_title(PGtkWindow(window),str2pchar(image.title));

    { **************** main vbox *************** }
    vbox:=gtk_vbox_new(false,0);
    gtk_container_add(GTK_CONTAINER(window),vbox);
    gtk_widget_show(vbox);

    { ******** Toolbar ******** }
    toolbar:=gtk_hbox_new(false,4);
    gtk_box_pack_start(GTK_BOX(vbox),toolbar,false,false,0);
    gtk_widget_show(toolbar);
    { **** open button **** }
    bopen:=siv_pixmap_button_new(xpm_fileopen);
    gtk_box_pack_start(GTK_BOX(toolbar),bopen,false,false,0);
    gtk_widget_show(bopen);
    g_signal_connect(PGTKOBJECT(bopen), 'clicked', GTK_SIGNAL_FUNC(@siv_visual_openfile), self);
    siv_tooltip_new(bopen,'Open file',nil);
    { **** reload button **** }
    breload:=siv_pixmap_button_new(xpm_reload);
    gtk_box_pack_start(GTK_BOX(toolbar),breload,false,false,0);
    gtk_widget_show(breload);
    g_signal_connect(PGTKOBJECT(breload), 'clicked', GTK_SIGNAL_FUNC(@siv_visual_reload), self);
    siv_tooltip_new(breload,'Reload window',nil);
    { **** clone button **** }
    bclone:=siv_pixmap_button_new(xpm_dupwin1);
    gtk_box_pack_start(GTK_BOX(toolbar),bclone,false,false,0);
    gtk_widget_show(bclone);
    g_signal_connect(PGTKOBJECT(bclone), 'clicked', GTK_SIGNAL_FUNC(@siv_visual_clone), self);
    siv_tooltip_new(bclone,'Clone window',nil);
    { **** fit button **** }
    bfit:=siv_pixmap_button_new(xpm_fit);
    gtk_box_pack_start(GTK_BOX(toolbar),bfit,false,false,0);
    gtk_widget_show(bfit);
    g_signal_connect(PGTKOBJECT(bfit), 'clicked', GTK_SIGNAL_FUNC(@siv_visual_fit), self);
    siv_tooltip_new(bfit,'Fit window',nil);
    { **** info button **** }
    info_window:=nil;
    binfo:=siv_pixmap_button_new(xpm_information);
    gtk_box_pack_start(GTK_BOX(toolbar),binfo,false,false,0);
    gtk_widget_show(binfo);
    g_signal_connect(PGTKOBJECT(binfo), 'clicked', GTK_SIGNAL_FUNC(@siv_visual_info), self);
    siv_tooltip_new(binfo,'Show information about image',nil);
    { **** history combo **** }
    histcombo:=gtk_combo_box_entry_new_text;
    if fn<>'' then gtk_combo_box_append_text(PGtkComboBox(histcombo),str2pchar(fn));
    for i:=0 to history.count-1 do
        if history.entries[i]<>fn then gtk_combo_box_append_text(PGtkComboBox(histcombo),str2pchar(history.entries[i]));
    gtk_combo_box_set_active(PGtkComboBox(histcombo),0);
    gtk_box_pack_start(GTK_BOX(toolbar),histcombo,true,true,0);
    gtk_widget_show(histcombo);
    siv_tooltip_new(histcombo,'File history',nil);
    gtk_combo_box_set_focus_on_click(PGtkComboBox(histcombo),false);
    g_signal_connect( PGtkObject(gtk_bin_get_child(PGtkBin(histcombo))), 'activate', GTK_SIGNAL_FUNC(@siv_visual_history_activate), self );

    { **** combinemte button **** }
    bcombinemte:=siv_pixmap_button_new(xpm_combine);
    gtk_box_pack_start(GTK_BOX(toolbar),bcombinemte,false,false,0);
    g_signal_connect(PGTKOBJECT(bcombinemte),'clicked',GTK_SIGNAL_FUNC(@siv_visual_combinemte),self);
    siv_tooltip_new(bcombinemte,'Combine multi time exposures (at loading)',nil);
    bnocombinemte:=siv_pixmap_button_new(xpm_nocombine);
    gtk_box_pack_start(GTK_BOX(toolbar),bnocombinemte,false,false,0);
    g_signal_connect(PGTKOBJECT(bnocombinemte),'clicked',GTK_SIGNAL_FUNC(@siv_visual_nocombinemte),self);
    siv_tooltip_new(bnocombinemte,'Don''t combine multi time exposures (at loading)',nil);
    combinemte:=cmte;
    if cmte then gtk_widget_show(bnocombinemte)
    else gtk_widget_show(bcombinemte);
    { **** quit button **** }
    bquit:=siv_pixmap_button_new(xpm_quit);
    gtk_box_pack_start(GTK_BOX(toolbar),bquit,false,false,0);
    gtk_widget_show(bquit);
    g_signal_connect_swapped(GTK_OBJECT(bquit),'clicked',GTK_SIGNAL_FUNC(@gtk_widget_destroy),GTK_OBJECT(window));
    siv_tooltip_new(bquit,'Quit',nil);

    { ******** drawing area ******** }
    marker_x:=-1; marker_y:=-1;
    iw:=CSIV_ImageWidget.init(window,image);
    gtk_widget_set_events(iw.draw,GDK_BUTTON_PRESS_MASK
        or GDK_KEY_PRESS_MASK
        or GDK_FOCUS_CHANGE_MASK
        or GDK_POINTER_MOTION_MASK
        or GDK_POINTER_MOTION_HINT_MASK);
    gtk_widget_set_flags(iw.draw,GTK_CAN_FOCUS);
    g_signal_connect(GTK_OBJECT(iw.draw),'button_press_event',TGtkSignalFunc(@siv_visual_marker_button_press),self);
    g_signal_connect(GTK_OBJECT(iw.draw),'motion_notify_event',TGtkSignalFunc(@siv_visual_marker_motion),self);
    g_signal_connect_after(GTK_OBJECT(iw.draw),'key_press_event',TGtkSignalFunc(@siv_visual_marker_key_press),self);
    g_signal_connect(GTK_OBJECT(iw.draw),'focus_in_event',TGtkSignalFunc(@siv_visual_marker_focus_in),self);
    g_signal_connect(GTK_OBJECT(iw.draw),'focus_out_event',TGtkSignalFunc(@siv_visual_marker_focus_out),self);
    gtk_box_pack_start(GTK_BOX(vbox),iw.widget,true,true,0);
    iw.show;

    { ******** info bar ******** }
    infobar:=gtk_hbox_new(false,0);
    gtk_box_pack_start(GTK_BOX(vbox),infobar,false,false,0);
    gtk_widget_show(infobar);

    { **** pointer **** }
    eventbox:=gtk_event_box_new;
    gtk_box_pack_start(GTK_BOX(infobar),eventbox,true,true,0);
    gtk_widget_show(eventbox);
    s:='Mouse: ';
    pointer_label:=gtk_label_new(str2pchar(s));
    gtk_label_set_justify(PGtkLabel(pointer_label),GTK_JUSTIFY_CENTER);
    fillchar(s[1],length(s),'_');
    gtk_label_set_pattern(PGtkLabel(pointer_label),str2pchar(s));
    gtk_container_add(PGtkContainer(eventbox),pointer_label);
    gtk_widget_show(pointer_label);
    siv_tooltip_new(eventbox,'Position of the mouse pointer',nil);

    { **** marker **** }
    marker_x:=-1; marker_y:=-1;
    eventbox:=gtk_event_box_new;
    gtk_box_pack_start(GTK_BOX(infobar),eventbox,true,true,0);
    gtk_widget_show(eventbox);
    s:='Marker: ';
    marker_label:=gtk_label_new(str2pchar(s));
    gtk_label_set_justify(PGtkLabel(marker_label),GTK_JUSTIFY_CENTER);
    fillchar(s[1],length(s),'_');
    gtk_label_set_pattern(PGtkLabel(marker_label),str2pchar(s));
    gtk_container_add(PGtkContainer(eventbox),marker_label);
    gtk_widget_show(marker_label);
    siv_tooltip_new(eventbox,'Position and color of the marked pixel',nil);

    { ******** show/hide control/history bar ******** }
    hbox1:=gtk_hbox_new(false,0);
    gtk_box_pack_start(GTK_BOX(vbox),hbox1,false,false,0);
    gtk_widget_show(hbox1);

    bshow_hb:=siv_pixmap_button_new(xpm_down_arrows);
    gtk_button_set_relief(PGtkButton(bshow_hb),gtk_relief_half);
    gtk_box_pack_start(GTK_BOX(hbox1),bshow_hb,true,true,0);
    g_signal_connect(PGTKOBJECT(bshow_hb), 'clicked', GTK_SIGNAL_FUNC(@siv_visual_show_histbar), self);
    siv_tooltip_new(bshow_hb,'Show histogram',nil);

    bhide_hb:=siv_pixmap_button_new(xpm_up_arrows);
    gtk_button_set_relief(PGtkButton(bhide_hb),gtk_relief_half);
    gtk_box_pack_start(GTK_BOX(hbox1),bhide_hb,true,true,0);
    g_signal_connect(PGTKOBJECT(bhide_hb), 'clicked', GTK_SIGNAL_FUNC(@siv_visual_hide_histbar), self);
    siv_tooltip_new(bhide_hb,'Hide histogram',nil);

    bshow_cb:=siv_pixmap_button_new(xpm_down_arrows);
    gtk_button_set_relief(PGtkButton(bshow_cb),gtk_relief_half);
    gtk_box_pack_start(GTK_BOX(hbox1),bshow_cb,true,true,0);
    g_signal_connect(PGTKOBJECT(bshow_cb), 'clicked', GTK_SIGNAL_FUNC(@siv_visual_show_controlbar), self);
    siv_tooltip_new(bshow_cb,'Show control bar',nil);

    bhide_cb:=siv_pixmap_button_new(xpm_up_arrows);
    gtk_button_set_relief(PGtkButton(bhide_cb),gtk_relief_half);
    gtk_box_pack_start(GTK_BOX(hbox1),bhide_cb,true,true,0);
    g_signal_connect(PGTKOBJECT(bhide_cb), 'clicked', GTK_SIGNAL_FUNC(@siv_visual_hide_controlbar), self);
    siv_tooltip_new(bhide_cb,'Hide control bar',nil);

    { ******** history bar ******** }
    histbar:=gtk_hbox_new(false,0);
    gtk_box_pack_start(GTK_BOX(vbox),histbar,false,false,0);
    if shb then
    begin
        gtk_widget_show(bhide_hb);
        gtk_widget_show(histbar);
    end
    else gtk_widget_show(bshow_hb);

    { **** histl **** }
    histl:=siv_draw_new(image.histlimg, siv_histl_xsize, siv_hist_ysize, nil, GTK_SHADOW_IN, 0);
    gtk_box_pack_start(GTK_BOX(histbar),histl,true,false,2);
    gtk_widget_show(histl);
    siv_tooltip_new(histl,'Histogram of the raw image',nil);

    { **** histr **** }
    tmpwid:=siv_pixmap_new(xpm_right_arrow);
    gtk_box_pack_start(GTK_BOX(histbar),tmpwid,false,false,0);
    gtk_widget_show(tmpwid);

    histr:=siv_draw_new(image.histrimg, siv_histr_xsize, siv_hist_ysize, nil, GTK_SHADOW_IN, 0);
    gtk_box_pack_start(GTK_BOX(histbar),histr,true,false,4);
    gtk_widget_show(histr);
    siv_tooltip_new(histr,'Histogram of the visible image',nil);

    { ******** control bar ******** }
    controlbar:=gtk_hbox_new(false,0);
    gtk_box_pack_start(GTK_BOX(vbox),controlbar,false,false,0);
    if scb then
    begin
        gtk_widget_show(bhide_cb);
        gtk_widget_show(controlbar);
    end
    else gtk_widget_show(bshow_cb);

    { **** left vbox **** }
    vboxl:=gtk_vbox_new(false,0);
    gtk_box_pack_start(GTK_BOX(controlbar),vboxl,true,true,4);
    gtk_widget_show(vboxl);

    { ** limits ** }
    vbox1:=gtk_vbox_new(false,0);
    gtk_box_pack_start(GTK_BOX(vboxl),vbox1,false,false,8);
    gtk_widget_show(vbox1);
    i:=siv_string_width(window,'Gamma=-0.00')+4;
    { * ul * }
    ul_scale:=CSIV_HScale.init(window, image.ul, 0, 65535, 8, 128);
    with ul_scale do
    begin
        connect(GTK_SIGNAL_FUNC(@siv_limits_update), self);
        set_tooltip('Change upper limit');
        set_label(i,'qu='+siv_real2str(image.qu*100,3)+'%');
        gtk_box_pack_start(GTK_BOX(vbox1),widget,false,false,0);
        show;
    end;
    { * ll * }
    ll_scale:=CSIV_HScale.init(window, image.ll, 0, 65535, 8, 128);
    with ll_scale do
    begin
        connect(GTK_SIGNAL_FUNC(@siv_limits_update), self);
        set_tooltip('Change lower limit');
        set_label(i,'ql='+siv_real2str(image.ql*100,3)+'%');
        gtk_box_pack_start(GTK_BOX(vbox1),widget,false,false,0);
        show;
    end;

    { ** gamma/beta ** }
    vbox1:=gtk_vbox_new(false,0);
    gtk_box_pack_start(GTK_BOX(vboxl),vbox1,false,false,0);
    gtk_widget_show(vbox1);
    { * gamma slider * }
    gam_scale:=CSIV_HScale.init(window, image.gam, -1, 1, 0.01, 0.05);
    with gam_scale do
    begin
        connect(GTK_SIGNAL_FUNC(@siv_bg_update), self);
        set_tooltip('Change gamma value');
        set_label(i,'Gamma='+siv_real2str(image.gam,2));
        gtk_box_pack_start(GTK_BOX(vbox1),widget,false,false,0);
        show;
    end;
    { * beta slider * }
    bet_scale:=CSIV_HScale.init(window, image.bet, -1, 1, 0.01, 0.05);
    with bet_scale do
    begin
        connect(GTK_SIGNAL_FUNC(@siv_bg_update), self);
        set_tooltip('Change beta value');
        set_label(i,'Beta='+siv_real2str(image.bet,2));
        gtk_box_pack_start(GTK_BOX(vbox1),widget,false,false,0);
        show;
    end;

    { **** right vbox **** }
    vboxr:=gtk_vbox_new(false,0);
    gtk_box_pack_start(GTK_BOX(controlbar),vboxr,false,false,4);
    gtk_widget_show(vboxr);

    { ** color menus ** }
    hbox1:=gtk_hbox_new(false,0);
    gtk_box_pack_start(GTK_BOX(vboxr),hbox1,false,false,8);
    gtk_widget_show(hbox1);

    { * red * }
    tmpwid:=gtk_drawing_area_new;
    gtk_widget_show(tmpwid);
    gtk_box_pack_start(GTK_BOX(hbox1),tmpwid,true,true,0);
    tmpwid:=siv_label_new('Red:');
    gtk_widget_show(tmpwid);
    gtk_box_pack_start(GTK_BOX(hbox1),tmpwid,false,false,0);
    red_menu:=create_color_menu(image.red);
    gtk_widget_show(red_menu);
    gtk_box_pack_start(GTK_BOX(hbox1),red_menu,false,false,0);
    siv_tooltip_new(red_menu,'Set red color',nil);
    { * green * }
    tmpwid:=gtk_drawing_area_new;
    gtk_widget_show(tmpwid);
    gtk_box_pack_start(GTK_BOX(hbox1),tmpwid,true,true,0);
    tmpwid:=siv_label_new('   Green:');
    gtk_widget_show(tmpwid);
    gtk_box_pack_start(GTK_BOX(hbox1),tmpwid,false,false,0);
    green_menu:=create_color_menu(image.green);
    gtk_widget_show(green_menu);
    gtk_box_pack_start(GTK_BOX(hbox1),green_menu,false,false,0);
    siv_tooltip_new(green_menu,'Set green color',nil);
    { * blue * }
    tmpwid:=gtk_drawing_area_new;
    gtk_widget_show(tmpwid);
    gtk_box_pack_start(GTK_BOX(hbox1),tmpwid,true,true,0);
    tmpwid:=siv_label_new('   Blue:');
    gtk_widget_show(tmpwid);
    gtk_box_pack_start(GTK_BOX(hbox1),tmpwid,false,false,0);
    blue_menu:=create_color_menu(image.blue);
    gtk_widget_show(blue_menu);
    gtk_box_pack_start(GTK_BOX(hbox1),blue_menu,false,false,0);
    siv_tooltip_new(blue_menu,'Set blue color',nil);
    tmpwid:=gtk_drawing_area_new;
    gtk_widget_show(tmpwid);
    gtk_box_pack_start(GTK_BOX(hbox1),tmpwid,true,true,0);

    { ** scale slider ** }
    scale_scale:=CSIV_HScale.init(window, -image.scale, -siv_maxscale, siv_maxscale, 1, 1);
    with scale_scale do
    begin
        connect(GTK_SIGNAL_FUNC(@siv_scale_update), self);
        set_tooltip('Zoom image');
        s:='Scale=10:0';
        i:=siv_string_width(window,str2pchar(s))+4;
        set_label(i,'Scale='+siv_scale2str(image.scale));
        gtk_box_pack_start(GTK_BOX(vboxr),widget,false,false,8);
        show;
    end;

    { ** toggle buttons ** }
    hbox1:=gtk_hbox_new(false,0);
    gtk_box_pack_start(GTK_BOX(vboxr),hbox1,false,false,4);
    gtk_widget_show(hbox1);
    { * inv * }
    inv_tb:=siv_toggle_button_new_with_label('Inverted');
    gtk_toggle_button_set_active(PGtkToggleButton(inv_tb),image.inv);
    gtk_box_pack_start(GTK_BOX(hbox1),inv_tb,true,true,8);
    g_signal_connect(PGTKOBJECT(inv_tb), 'toggled', GTK_SIGNAL_FUNC(@siv_inv_update), self);
    gtk_widget_show(inv_tb);
    siv_tooltip_new(inv_tb,'Invert image',nil);
    { * swap * }
    swap_b:=siv_button_new_with_label('Swap');
    gtk_box_pack_start(GTK_BOX(hbox1),swap_b,true,true,4);
    g_signal_connect(PGTKOBJECT(swap_b), 'clicked', GTK_SIGNAL_FUNC(@siv_swap_update), self);
    gtk_widget_show(swap_b);
    siv_tooltip_new(swap_b,'Swap limits',nil);
    { * log * }
    log_tb:=siv_toggle_button_new_with_label('Logarithmic');
    gtk_toggle_button_set_active(PGtkToggleButton(log_tb),image.log);
    gtk_box_pack_end(GTK_BOX(hbox1),log_tb,true,true,4);
    g_signal_connect(PGTKOBJECT(log_tb), 'toggled', GTK_SIGNAL_FUNC(@siv_log_update), self);
    gtk_widget_show(log_tb);
    siv_tooltip_new(log_tb,'Logarithmic view for histograms',nil);

    gtk_widget_grab_focus(iw.draw);
    if anc=nil then fit
    else begin
//        gtk_window_get_size(PGtkWindow(anc.window), @rect.x, @rect.y);
//        gtk_window_resize(PGtkWindow(window), rect.x, rect.y);
        siv_adjustment_copy_from(iw.hadj, anc.iw.hadj);
        siv_adjustment_copy_from(iw.vadj, anc.iw.vadj);
        if (anc.marker_x>=0) and (anc.marker_y>0) then setmarker(anc.marker_x,anc.marker_y);
    end;
    next:=nil;
    tmp:=@siv_visual_list;
    while tmp^<>nil do
        tmp:=@tmp^.next;
    tmp^:=self;
    siv_debug_end;
end;

{ ******** done *********************************************************** }
destructor CSIV_Visual.done;
begin
    siv_debug_begin('CSIV_Visual.done');
    if (info_window<>nil) and (gtk_is_window(info_window)) then gtk_widget_destroy(info_window);
    history.destroy;
    iw.done;
    ll_scale.done;
    ul_scale.done;
    bet_scale.done;
    gam_scale.done;
    scale_scale.done;
    image.done;
    siv_debug_end;
end;


{ ******** redraw ********************************************************* }
procedure CSIV_Visual.redraw;
begin
    image.update;
    iw.redraw;
    if image.redraw_histl then siv_draw_redraw(PSivDraw(histl));
    if image.redraw_histr then siv_draw_redraw(PSivDraw(histr));
    GTK_CONTAINER_CHECK_RESIZE(PGtkContainer(window));
end;


{ ******** fit ************************************************************ }
procedure CSIV_Visual.fit;
var requisition  : TGtkRequisition;
    rect         : TGDKRectAngle;
begin
    // calculate maximum window size
    gdk_window_get_frame_extents(window^.window, @rect);
    gtk_window_get_size(PGtkWindow(window), @rect.x, @rect.y);
    rect.width-=rect.x; rect.height-=rect.y;
    gdk_drawable_get_size(PGdkDrawable(gdk_screen_get_root_window(gtk_window_get_screen(PGtkWindow(window)))), @rect.x, @rect.y);
    rect.x-=rect.width+100; rect.y-=rect.height+100;

    //calculate window size
    iw.enable_fit_request;
    gtk_widget_queue_resize(iw.widget);
    gtk_widget_size_request(vbox,@requisition);
    if requisition.width>rect.x then requisition.width:=rect.x;
    if requisition.height>rect.y then requisition.height:=rect.y;
    gtk_widget_queue_resize(iw.widget);
    gtk_window_resize(PGtkWindow(window), requisition.width, requisition.height);
    //gtk_widget_queue_resize(window);
end;

{ ******** opennew ******************************************************** }
procedure CSIV_Visual.opennew(fn:simgstring);
var fi      : TSIMG_FPImage;
    h       : CSIV_History;
    scb,shb : boolean;
begin
    siv_debug_begin('CSIV_Visual.opennew: fn='+fn);
    try
        SImg_ReadImageFile(siv_config,fn,fi);
        if SImg_Recover(siv_config, fi) then
        begin
//            if (file_selection<>nil) and gtk_is_file_selection(file_selection) then gtk_widget_destroy(file_selection);
            h:=history;
            scb:=gtk_widget_visible(controlbar);
            shb:=gtk_widget_visible(histbar);
            CSIV_Visual.init(siv_load_fi, fi, fn, '', h, scb,shb,combinemte, nil);
            SImg_Free(siv_config, fi);
            if not SImg_Recover(siv_config,image.img16) then
            begin
                h:=nil;
                siv_dont_stop:=true;
                gtk_widget_destroy(window);
            end;
        end;
    except
        SIV_OutputCatchedException;
    end;
    siv_debug_end;
end;


{ ******** setmarker ****************************************************** }
procedure CSIV_Visual.setmarker(x,y:longint);
var i  : longint;
    s  : simgstring;
begin
    if (image.img16.nx<1) or (image.img16.ny<1) or (image.img16.nc<1) or (image.img16.pt=nil) then exit;
    if x<0 then x:=0;
    if y<0 then y:=0;
    if image.scale>0 then i:=image.img16.nx mod (image.scale+1)
    else i:=0;
    if x>=image.img16.nx-i then x:=image.img16.nx-i-1;
    if image.scale>0 then i:=image.img16.ny mod (image.scale+1);
    if y>=image.img16.ny-i then y:=image.img16.ny-i-1;
    if (x<>marker_x) or (y<>marker_y) then
    begin
        if image.scale<0 then siv_draw_marker_move(PSIVDraw(iw.draw), x*(1-image.scale)+(1-image.scale) shr 1, y*(1-image.scale)+(1-image.scale) shr 1, true)
        else if image.scale=0 then siv_draw_marker_move(PSIVDraw(iw.draw),x,y, true)
        else siv_draw_marker_move(PSIVDraw(iw.draw), x div (1+image.scale) ,y div (1+image.scale), true);
        marker_x:=x;
        marker_y:=y;
        if gtk_widget_visible(histbar) then
        begin
            image.history_markers(x,y);
            siv_draw_redraw(PSivDraw(histl));
            siv_draw_redraw(PSivDraw(histr));
        end;
        s:='Marker: '+siv_pos2str(x,y)+'  '+siv_poscolor2str(image,x,y);
        gtk_label_set_text(PGtkLabel(marker_label),str2pchar(s));
        if not gtk_widget_has_focus(iw.draw) then gtk_widget_grab_focus(iw.draw);
    end;
end;

{ ******** CSIV_Visual.create_color_menu ********************************** }
function CSIV_Visual.create_color_menu(default:longint):PGtkWidget;
var i     : longint;
    s     : shortstring;
    combo : PGtkWidget;

begin
    combo:=gtk_combo_box_new_text;
    for i:=0 to image.img16.nc-1 do
    begin
        str(i,s);
        gtk_combo_box_append_text(PGtkComboBox(combo), str2pchar(s));
    end;
    gtk_combo_box_set_focus_on_click(PGtkComboBox(combo),false);
    gtk_combo_box_set_active(PGtkComboBox(combo),default);
    g_signal_connect( PGtkObject(combo), 'changed', GTK_SIGNAL_FUNC(@siv_color_update), self );
    result:=combo;
end;
